# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Flips metric goal and value signs."""
from typing import Sequence
from vizier import pyvizier
from vizier._src.benchmarks.experimenters import experimenter as experimenter_lib


class SignFlipExperimenter(experimenter_lib.Experimenter):
  """Flips the goal and objective values of the wrapped Experimenter."""

  def __init__(
      self,
      experimenter: experimenter_lib.Experimenter,
      flip_objectives_only: bool = True,
  ):
    """Init.

    Args:
      experimenter: Original experimenter to flip.
      flip_objectives_only: Whether to only flip objective metrics in trials
        corresponding to problem statement objectives. If True, auxiliary
        metrics will not be flipped.
    """
    self._experimenter = experimenter
    self._flip_objectives_only = flip_objectives_only
    self._original_objectives = {
        m_config.name: m_config
        for m_config in self._experimenter.problem_statement().metric_information
    }

  def problem_statement(self) -> pyvizier.ProblemStatement:
    problem = self._experimenter.problem_statement()
    for metric_config in problem.metric_information:
      if metric_config.goal.is_maximize:
        metric_config.goal = pyvizier.ObjectiveMetricGoal.MINIMIZE
      elif metric_config.goal.is_minimize:
        metric_config.goal = pyvizier.ObjectiveMetricGoal.MAXIMIZE
    return problem

  def evaluate(self, suggestions: Sequence[pyvizier.Trial]):
    self._experimenter.evaluate(suggestions)
    for suggestion in suggestions:
      if suggestion.final_measurement is None:
        continue

      metric_dict = {}
      for name, metric in suggestion.final_measurement.metrics.items():
        if self._flip_objectives_only and name in self._original_objectives:
          metric_dict[name] = pyvizier.Metric(value=-1.0 * metric.value)
        elif not self._flip_objectives_only:
          metric_dict[name] = pyvizier.Metric(value=-1.0 * metric.value)
        else:
          metric_dict[name] = metric
      suggestion.final_measurement.metrics = metric_dict



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Common classes shared between Study and Trial."""

import collections
from collections import abc
from typing import DefaultDict, Dict, overload, Iterator
from typing import Iterable, List, Optional, Tuple, TypeVar, Union, Type

from absl import logging
import attr

from google.protobuf import any_pb2
from google.protobuf.message import Message

_M = TypeVar('_M', bound=Message)
T = TypeVar('T')
T1 = TypeVar('T1')
T2 = TypeVar('T2')
MetadataValue = Union[str, any_pb2.Any, Message]
_V = TypeVar('_V', bound=MetadataValue)

# Namespace Encoding.
#
# By definition, ∀ ns ∈ Namespace, Namespace.decode(ns.encode()) == ns.
# The tricky part of that definition is handling namespaces with components
# that are empty strings.  Notably, we want to make sure that
# Namespace(()).encode() != Namespace(('',)).encode().
# So, we set up the mapping:
# Namespace(()).encode() -> ''
# Namespace((s,)).encode() -> ':s'
# Namespace((s, s)).encode() -> ':s:s',
# et cetera, and note that every tuple gets a unique encoding, even if $s is the
# empty string.  (As long as we escape colons properly.)
#
# So, ns.encode() is a bijection, therefore it has an inverse which we call
# Namespace.decode(s).


def _parse(arg: str) -> Tuple[str, ...]:
  """Parses an encoded namespace string into a namespace tuple."""
  # The tricky part here is that arg.split('') has a length of 1, so it can't
  # generate a zero-length tuple; we handle that corner case manually.
  if not arg:
    return ()
  # And, then, once we've handled the case of _parse(''), we note that all the
  # other encoded strings begin with a colon.  It thus contains no information
  # and we can remove it.
  # TODO: Once we're on Python 3.9, use: arg = arg.removeprefix(':')
  if arg.startswith(':'):
    arg = arg[1:]
  # The rest of the algorithm is that we split on all colons, both
  # escaped and unescaped.  Then, we walk through the list of fragments and
  # join back together the colons that were preceeded by an escape character,
  # dropping the escape character as we go.
  fragments = arg.split(':')
  output = []
  join = False
  for frag in fragments:
    if join and frag and frag[-1] == '\\':
      output[-1] += ':' + frag[:-1]
      join = True
    elif join:  # Doesn't end in an escape character.
      output[-1] += ':' + frag
      join = False
    elif frag and frag[-1] == '\\':  # Don't join to previous.
      output.append(frag[:-1])
      join = True
    else:  # Don't join to previous and doesn't end in an escape.
      output.append(frag)
      join = False
  return tuple(output)


@attr.frozen(eq=True, order=True, hash=True, auto_attribs=True, init=False)
class Namespace(abc.Sequence):
  r"""A namespace for the Metadata class.

  Namespaces represent a tree of Metadata; each namespace object can be thought
  of as a tuple of components obtained by walking the tree from the root.
  This makes it easy to give each part of your algorithm its own namespace,
  to avoid name collisions.  E.g. if your algorithm A uses sub-algorithms B and
  C, you might have namespaces ":A", ":A:B", and ":A:C".

  NOTE: The empty namespace is writeable by users via a RPC in Vizier's
    user-facing API; other namespaces are writeable only by Pythia algorithms.
    (Users can read all namespaces.) So, to minimize collisions, please avoid
    the empty namespace unless your algorithm needs to read user data.

  You can create a Namespace from a tuple of strings, e.g.
  Namespace(('a', 'b')).  Or, you can create a Namespace from a single string
  with Namespace.decode(s); this parses the string into components, splitting at
  colons.  For instance Namespace.decode(':a:b') gives you a two-component
  namespace, equivalent to Namespace(('a', 'b')).
  (Note that in the tuple case, the strings are not parsed and colons are
  treated as ordinary characters.)

  TLDR: If you decode() a namespace from a string, then ":" is a
    reserved character, but when constructing from a tuple, there are no
    reserved characters.

  Decoding the string form:
  * Initial colons don't matter: Namespace.decode(':a') == Namespace('a');
    this is a single-component namespace.
  * Colons separate components:
    Namespace.decode('a:b') == Namespace(('a', 'b')).
    (This is a two-component namespace.)
  * Colons are encoded as r'\:':
    Namespace.decode('a\\:b') == Namespace(('a:b',)).
    (This is a single-component namespace.)

  Conversions: For a Namespace x,
  * Namespace.decode(x.encode()) == x; here, x.encode() will be a string with
    colons separating the components.
  * Namespaces act as a Sequence[str], so Namespace(tuple(x)) == x and
    Namespace(x) == x.
  """

  _as_tuple: Tuple[str, ...] = attr.field(hash=True, eq=True, order=True)

  def __init__(self, arg: Iterable[str] = ()):
    """Generates a Namespace from its component strings.

    Args:
      arg: typically, a tuple of strings.
    """
    arg = tuple(arg)
    self.__attrs_init__(as_tuple=arg)

  _ns_repr_table = str.maketrans({':': r'\:'})

  @classmethod
  def decode(cls, s: str) -> 'Namespace':
    r"""Decode a string into a Namespace.

    For a Namespace x, Namespace.decode(x.encode()) == x.

    Args:
      s: A string where ':' separates namespace components, and colon is escaped
        as r'\:'.

    Returns:
      A namespace.
    """
    return Namespace(_parse(s))

  def encode(self) -> str:
    """Encodes a Namespace into a string.

    Given a Namespace x, Namespace.decode(x.encode()) == x.

    Returns:
      Colons are escaped, then Namespace components are joined by colons.
    """
    return ''.join(
        [':' + c.translate(self._ns_repr_table) for c in self._as_tuple]
    )

  def __len__(self) -> int:
    """Number of components (elements of the tuple form)."""
    return len(self._as_tuple)

  def __add__(self, other: Iterable[str]) -> 'Namespace':
    """Appends components onto the namespace."""
    return Namespace(self._as_tuple + tuple(other))

  @overload
  def __getitem__(self, key: int) -> str:
    ...

  @overload
  def __getitem__(self, key: slice) -> 'Namespace':
    ...

  def __getitem__(self, key):
    """Retrieves item by the specified key."""
    if isinstance(key, int):
      return self._as_tuple[key]
    return Namespace(self._as_tuple[key])

  def __str__(self) -> str:
    """Shows the namespace, fully escaped."""
    return self.encode()

  def __repr__(self) -> str:
    """Shows the namespace, fully escaped."""
    return f'Namespace({self.encode()})'

  def startswith(self, prefix: Iterable[str]) -> bool:
    """Returns True if this namespace starts with $prefix.

    So, if the current namespace is "a:b:c", then startswith() will return True
    when called with prefix=(), ('a',), ('a', 'b'), and ('a', 'b', 'c');
    otherwise False.

    Args:
      prefix: namespace components or a Namespace object.

    Returns:
    """
    ns_prefix = Namespace(prefix)
    return self[: len(ns_prefix)] == ns_prefix


class _MetadataSingleNameSpace(Dict[str, MetadataValue]):
  """Stores metadata associated with one namespace."""

  pass


class Metadata(abc.MutableMapping):
  """Metadata class: a key-value dict-like mapping, with namespaces.

  This is the main interface for reading metadata from a Trial or StudyConfig,
  or adding metadata to a Trial/StudyConfig.  Loosely speaking, within each
  namespace, Metadata acts like a dictionary mapping from a string key to a
  string or protobuf. (See more about namespaces below.)

  Metadata can be initialized from a dictionary:
    mm = Metadata({'foo': 'Foo'})
  And items can be retrieved with:
    mm.get('foo')  # Returns 'Foo'
    mm['foo']      # Returns 'Foo'

  More items can be added with:
    mm = Metadata({'foo': 'Foo'})
    mm['bar'] = 'Bar'  # Add a single item
    mm.update({'a': 'A'}, gleep='Gleep')  # Add two items

  By default, items are added to the root/empty namespace.
  Vizier users can only add metadata to the empty namespace (the Vizier service
  will reject attempts by users to add metadata elsewhere); Pythia algorithms
  can add metadata to any namespace, but should normally work in a single unique
  namespace, and should avoid the root namespace, unless they intend to
  pass data to/from Vizier users.

  1. Keys are namespaced. Each Metadata object only interacts with one
    Namespace.

    Namespaces form a tree, and you can walk down the tree.

    NOTE ns(s: str) takes one step down the namespace tree. For nearly all
    practical purposes, ignore abs_ns() which is only used for conversions
    to and from protobufs.

    mm = Metadata({'foo': 'foofoo'})
    # $mm is created with its current namespace equal to the root/empty
    # namespace.
    mm.ns('NewName')['bar'] = 'Bar'
    # We've added an item in the ":NewName" namespace, but $mm's current
    # namespace is unchanged.

  2. Values can be protobufs. If `metadata['foo']` is an instance of `MyProto`
    proto message or an `Any` proto that packs a `MyProto` message, then the
    proto can be recovered by calling:

    my_proto = metadata.get('foo', cls=MyProto)
    isinstance(my_proto, MyProto) # Returns `True`

    NOTE that the bracket operator doesn't work well for protobufs:
      metadata['foo'] will return an `Any` protobuf instead of a MyProto.
      For protos, you may wish to use
      `metadata.get_or_error('foo', cls=MyProto)` instead of the bracket form.

  3. An iteration over a Metadata object only shows you the data in the current
    namespace.  So,

    mm = Metadata({'foo': 'foofoo'})
    for k, v in mm.ns('gleep'):
      ...

    will not yield anything because there are no keys in the 'gleep' namespace.
    WARNING: Because of this behavior, if you iterate over Metadata(mm), you
      will quietly drop metadata from all but mm's current namespace.
    NOTE also that the type of $v is MetadataValue, which can carry strings and
      protos; you may want to use mm.get_or_error(key, cls=proto_class) to
      unpack the contained proto.

  4. To iterate over all the keys in all the namespaces use

    mm = Metadata()
    mm.ns('gleep')['x'] = 'X'
    for ns, k, v in mm.all_items():
      # iteration will include ('gleep', 'x', 'X')
      # Be aware that type(v) is MetadataValue, which can carry either strings
      # or protos.
  """

  def __init__(
      self,
      *args: Union[
          Dict[str, MetadataValue], Iterable[Tuple[str, MetadataValue]]
      ],
      **kwargs: MetadataValue,
  ):
    """Construct; this follows dict(), and puts data in the root namespace.

    You can pass it a dict, or an object that yields (key, value)
    pairs, and those pairs will be put in the root namespace.

    Args:
      *args: A dict or an iterable the yields key-value pairs.
      **kwargs: key=value pairs to be added to the specified namespace.
    """
    self._stores: DefaultDict[Namespace, _MetadataSingleNameSpace] = (
        collections.defaultdict(_MetadataSingleNameSpace)
    )
    self._namespace = Namespace()
    self._store = self._stores[self._namespace]
    self._store.update(*args, **kwargs)

  def ns(self, component: str) -> 'Metadata':
    r"""Switches to a deeper namespace by appending one component.

    The entire tree of metadata is shared between $self and the returned value,
    but the returned value will have a deeper current namespace.  ($self is not
    modified.)

    Args:
      component: one component to be appended to the current namespace.

    Returns:
      A new Metadata object in the specified namespace; the new object shares
      metadata with $self.
    """
    new_ns: Namespace = self._namespace + (component,)
    return self._copy_core(new_ns)

  def __repr__(self) -> str:
    """Prints items in all namespaces."""
    itemlist: List[str] = []
    for namespace, store in self._stores.items():
      item_string = f'(namespace:{namespace}, items: {store})'
      itemlist.append(item_string)
    return 'Metadata({}, current_namespace={})'.format(
        ', '.join(itemlist), self._namespace.encode()
    )

  def __str__(self) -> str:
    """Prints items in the current namespace."""
    return 'namespace: {} items: {}'.format(str(self._namespace), self._store)

  def get_proto(self, key: str, *, cls: Type[_M]) -> Optional[_M]:
    """Deprecated: use get() instead."""
    logging.warning(
        'Metadata.get_proto() is deprecated, prefer Metadata.get().'
    )
    value = self._store.get(key, None)
    if value is None:
      return None

    if isinstance(value, cls):
      # Starting from 3.10, pytype supports typeguard, which obsoletes
      # the need for the `pytype:disable` clause.
      return value  # pytype: disable=bad-return-type
    if isinstance(value, any_pb2.Any):
      # `value` is an Any proto potentially packing `cls`.
      message = cls()
      success = value.Unpack(message)
      return message if success else None
    return None

  def get_or_error(self, key: str, *, cls: Type[T] = str) -> T:
    """Gets the metadata as type `cls`, or raises a KeyError.

    This acts like the square bracket operator, except that
    it lets you specify a class; it gets the metadata from the current
    namespace.

    Examples with string metadata:
      metadata = common.Metadata({'key': 'value'})
      assert metadata.get_or_error('key') == 'value'
      metadata.get_or_error('badkey')  # raises KeyError

    Examples with numeric values:
      metadata = common.Metadata({'float': '1.2', 'int': '60'})
      assert metadata.get_or_error('int', cls=int) == 60
      assert metadata.get_or_error('float', cls=float) == 1.2
      metadata.get_or_error('badkey', cls=float)      # raises KeyError

    Example with `Duration` and `Any` proto values:
      duration = Duration(seconds=60)
      anyproto = Any()
      anyproto.Pack(duration)
      metadata = common.Metadata({'duration': duration, 'any': anyproto})
      assert metadata.get_or_error('duration', cls=Duration) == duration
      assert metadata.get_or_error('any', cls=Duration).seconds == 60

    Args:
      key:
      cls: Desired type of the value.

    Returns:
      The matching metadata value is parsed into type `cls`. For proto messages,
      it involves unpacking an Any proto.

    Raises:
      KeyError if the metadata item is not present.
      TypeError or other errors if the string can't be converted to $cls.
    """
    value = self._store[key]
    if isinstance(value, cls):
      # Starting from 3.10, pytype supports typeguard, which obsoletes
      # the need for the `pytype:disable` clause.
      return value  # pytype: disable=bad-return-type
    elif isinstance(value, any_pb2.Any):
      # `value` is an Any proto potentially packing `cls`.
      message = cls()
      if not value.Unpack(message):
        logging.warning(
            'Cannot unpack message to %s: %s', cls, str(value)[:100]
        )
        raise TypeError('Cannot unpack to %s' % cls)
      return message
    else:
      return cls(value)

  def get(
      self, key: str, default: T1 = None, *, cls: Type[T2] = str
  ) -> Union[T1, T2]:
    """Gets the metadata as type `cls`, or $default if not present.

    This returns $default if the specified metadata item is not found.
    Note that there's always a default value, and the $default defaults to None.
    This gets the data from the current namespace.

    For string values, this function behaves exactly like a
    regular string-to-string dict (within its namespace).
      metadata = common.Metadata({'key': 'value'})
      metadata.get('key')  # returns 'value'
      metadata.get('badkey')  # returns None
      assert metadata.get('badkey', 'badvalue') == 'badvalue'

    Examples with numeric values:
      metadata = common.Metadata({'float': '1.2', 'int': '60'})
      value = metadata.get('int', cls=int)
      if value is not None:
        assert value == 60
      #
      metadata.get('float', cls=float)       # returns 1.2
      metadata.get('badkey', cls=float)      # returns None
      metadata.get('int', cls=int)           # returns 60
      assert metadata.get('float', 0.0, cls=float) == 1.2
      assert metadata.get('badkey', 1, cls=int) == 1
      assert metadata.get('badkey', 0.2, cls=float) == 0.2

    Example with `Duration` and `Any` proto values:
      duration = Duration(seconds=60)
      anyproto = Any()
      anyproto.Pack(duration)
      metadata = common.Metadata({'duration': duration, 'any': anyproto})
      duration_out =  metadata.get('duration', cls=Duration)
      if duration_out is not None:
        assert duration_out == duration
      any_out =  metadata.get('any', cls=Duration)
      if any_out is not None:
        assert any_out == duration

    Args:
      key:
      default: Default value.
      cls: Desired type of the value.

    Returns:
      $default if the key does not exist. Otherwise, the matching value is
      parsed into type `cls`. For proto messages, it involves unpacking an
      Any proto.

    Raises:
      TypeError or other errors if the string can't be converted to $cls.
    """
    try:
      return self.get_or_error(key, cls=cls)
    except KeyError:
      return default

  # TODO: Rename to `abs_namespaces`
  def namespaces(self) -> List[Namespace]:
    """List all namespaces for which there is at least one key."""
    return [ns for ns, store in self._stores.items() if store]

  # TODO: Rename to `namespaces`
  def subnamespaces(self) -> Tuple[Namespace, ...]:
    """Returns relative namespaces that are at or below the current namespace.

    For all `ns` in the returned value, `self.abs_ns(md.current_ns() + ns)` is
    not empty.
    # Examples:
    md = Metadata()
    md.ns('foo').ns('bar')['A'] = 'b'
    md.subnamespaces() == (Namespace(['foo', 'bar']),)
    md.ns('foo').subnamespaces() == (Namespace(['bar']),)

    Returns:
      For all namespaces that begin with the current namespace and are
      non-empty, this returns a namespace object that contains the relative
      path from the current namespace.
    """
    return tuple(
        [
            Namespace(ns[len(self._namespace) :])
            for ns, store in self._stores.items()
            if store and ns.startswith(self._namespace)
        ]
    )

  def current_ns(self) -> Namespace:
    """Displays the object's current Namespace."""
    return self._namespace

  def all_items(self) -> Iterator[Tuple[Namespace, str, MetadataValue]]:
    """Yields an iterator that walks through all metadata items.

    This iterates through all the metadata items in all namespaces, vs.
    __iter__() which just iterates through all the items in the current
    namespace.

    Yields:
      Tuple of (namespace, key, value).
    """
    for ns in self.namespaces():
      for k, v in self.abs_ns(ns).items():
        yield (ns, k, v)

  def items_by_cls(self, *, cls: Type[_V]) -> Iterator[Tuple[str, _V]]:
    """Yields an iterator over items whose type=$cls in the current namespace.

    This iterates through the metadata items in the current namespace, like
    __iter__(), except that it only returns items of the specified type.

    Args:
      cls: What type of objects to filter for?

    Yields:
      Tuple of (key, value).
    """
    for k_v in self.items():
      if isinstance(k_v[1], cls):
        yield k_v

  # START OF abstract methods inherited from `MutableMapping` base class.
  def __getitem__(self, key: str) -> MetadataValue:
    return self._store.__getitem__(key)

  def __setitem__(self, key: str, value: MetadataValue):
    self._store[key] = value

  def __delitem__(self, key: str):
    del self._store[key]

  def __iter__(self):
    return iter(self._store)

  def __len__(self):
    """The number of elements in the current namespace."""
    return len(self._store)

  def __bool__(self):
    """True if this instance contains any metadata in _any_ namespace."""
    for s in self._stores.values():
      if s:
        return True
    return False

  def __copy__(self) -> 'Metadata':
    """Shallow copy -- metadata continues to be shared.

    Returns:
      A copy of the object.
    """
    return self._copy_core(self._namespace)

  # END OF Abstract methods inherited from `MutableMapping` base class.

  def abs_ns(self, namespace: Iterable[str] = ()) -> 'Metadata':
    """Returns a metadata object set to the specified absolute namespace.

    NOTE Prefer using ns() instead in most cases.

    abs_ns() jumps to the root namespace and
    abs_ns(ns) jumps to the specified Namespace.

    (NOTE: ns() and abs_ns() take different argument types!)
    (NOTE: Neither ns() nor abs_ns() modify the Metadata object they are called
     on: they return a shallow copy that shares all metadata items, but
     which displays a different namespace.)

    # Use of abs_ns().
    mm.abs_ns(['NewName'])  # returns 'Bar'
    mmx = mm.ns('x')
    mmx.abs_ns(['NewName'])  # returns 'Bar2'
    mmx.abs_ns().get('foo')  # returns 'foofoo'

    # Multi-component namespaces.
    mm = Metadata()
    mm.ns('a').ns('b')['foo'] = 'AB-foo'
    mm.ns('a')['foo'] = 'A-foo'
    mm['foo']          # Throws a KeyError
    mm.ns('a')['foo']  # returns 'A-foo'
    mm.ns('a').ns('b')['foo']  # returns 'AB-foo'
    # abs_ns() can be also used:
    mm.abs_ns(['a', 'b']).get('foo')  # Returns 'ab-foo'
    mm.abs_ns(Namespace.decode('a:b')).get('foo')  # Returns 'ab-foo'

    All the Metadata object's data is shared between $self and the returned
    object, but the new Metadata object will have a different current
    namespace.  (Note that $self is not modified, and the current namespace of
    $self doesn't matter.)

    NOTE: $namespace can be a Namespace object, because you can iterate over
      a Namespace to get strings.

    Args:
      namespace: a list of Namespace components.  (Defaults to the root, empty
        Namespace.)

    Returns:
      A new Metadata object that shares data with $self, but the current
      namespace is one level deeper.
    """
    if isinstance(namespace, str):
      raise ValueError(
          'Passing str to abs_ns() is rarely intended and therefore '
          'considered an error. Carefully read the class doc and prefer '
          'using ns(). If you do decide abs_ns() is the right method, '
          'expclitily pass abs_ns([namespace]).'
      )
    return self._copy_core(Namespace(namespace))

  def _copy_core(self, ns: Namespace) -> 'Metadata':
    """Shallow copy: metadata is shared, default namespace changes.

    Args:
      ns: the namespace to use for the new object.

    Returns:
      A copy of the object.
    """
    md = Metadata()
    md._namespace = ns  # pylint: disable='protected-access'
    md._stores = self._stores  # pylint: disable='protected-access'
    md._store = md._stores[md._namespace]  # pylint: disable='protected-access'
    return md

  def update(
      self,
      *args: Union[
          Dict[str, MetadataValue], Iterable[Tuple[str, MetadataValue]]
      ],
      **kwargs: MetadataValue,
  ) -> None:
    self._store.update(*args, **kwargs)

  def attach(self, other: 'Metadata') -> None:
    """Attach the $other metadata as a descendent of this metadata.

    More precisely, it takes the part of `other`'s namespace that is at or
    below `other`'s current namespace, and attaches it to `self`'s current
    namespace.
    * Tree structure is preserved and nothing is flattened.
    * Attached data overwrites existing data, item-by-item, not
      namepace-by-namespace.

    So, if we have
    other = Metadata()
    other.abs_ns(('x', 'y', 'z'))['foo'] = 'bar'
    m = Metadata()
    m.ns('w').attach(other.ns('x'))
    then
    m.abs_ns(('w', 'y', 'z'))['foo'] will contain 'bar'.

    Args:
      other: a Metadata object to copy from.
    """
    for ns in other.subnamespaces():
      self._stores[self._namespace + ns].update(
          other.abs_ns(other.current_ns() + ns)
      )



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Classes for specifying Vizier algorithms."""
import pyglove as pg


class PseudoAlgorithm(pg.DNAGenerator):
  """Base class for algorithms that delegate actual logics to others."""

  def _setup(self) -> None:
    raise RuntimeError(f'{self!r} should be used with \'vizier\' backend.')

  def _propose(self) -> pg.DNA:
    self._should_never_trigger('propose')
    return pg.DNA(0)

  def _feedback(self, unused_dna: pg.DNA, unused_reward: float) -> None:
    self._should_never_trigger('feedback')

  def _should_never_trigger(self, method_name):
    raise RuntimeError(
        f'`{method_name}` of pseudo algorithm {self!r} should never be called.')


@pg.members(
    [(
        'name',
        pg.typing.Enum(
            'DEFAULT',
            [
                'DEFAULT',
                # Algorithms shared across Vizier platforms.
                'GRID_SEARCH',
                'RANDOM_SEARCH',
                # Google-Vizier only algorithms.
                # TODO: copybara away these algorithms for OSS.
                'CMAES',
                'GP_BANDIT',
                'LINEAR_COMBINATION_SEARCH',
                'HYPER_LCS',
                # OSS-Vizier only algorithms:
                'BOCS',
                'CMA_ES',
                'EMUKIT_GP_EI',
                'HARMONICA',
                'NSGA2',
                'QUASI_RANDOM_SEARCH',
            ]),
        'Name of Vizier predefined algorithm.')],
    metadata={'init_arg_list': ['name']})  # pylint: disable=bad-continuation
class BuiltinAlgorithm(PseudoAlgorithm):
  """Vizier built-in algorithm."""

  @property
  def name(self) -> str:
    return self.sym_init_args.name

  @property
  def multi_objective(self) -> bool:
    return self.name in ('DEFAULT', 'GP_BANDIT', 'LINEAR_COMBINATION_SEARCH',
                         'RANDOM_SEARCH')


# TODO: implement the details according to Vizier early stopping policy
# later. Move to stopping.py
class BuiltinEarlyStoppingPolicy(pg.tuning.EarlyStoppingPolicy):
  """Vizier built-in early stopping policy."""



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Simple regret score to compare algorithms."""

from typing import Union

import numpy as np
from scipy import stats
from vizier import pyvizier as vz


def t_test_mean_score(baseline_mean_values: Union[list[float], np.ndarray],
                      candidate_mean_values: Union[list[float], np.ndarray],
                      objective_goal: vz.ObjectiveMetricGoal) -> float:
  """Computes the one-sided T-test score.

  In case of a maximization (minimizatoin) problem, it scores the confidence
  that the mean of 'baseline_mean_values' is less (greater) than the mean of
  'candidate_mean_values'.

  The lower the score the higher the confidence that it's the case.

  One-sample
  ----------
  The test assumes t-distribution with mean given by 'candidate' and compute
  the probability of observing sample mean of 'baseline' or less.

  Two-sample
  ----------
  The test assumes 'baseline' and 'candidate' have the same mean and computes
  the probability that the 'baseline' sample mean is less than the 'candidate'
  sample mean.

  The lower the T-test p-value score the more confidence we have that
  'candidate' is indeed "better" than 'baseline'.

  Arguments:
    baseline_mean_values: List of baseline simple regret values.
    candidate_mean_values: List of candidate simple regret values.
    objective_goal: The optimization problem type (MAXIMIZE or MINIMIZE).

  Returns:
    The p-value score of the one-sided T test.
  """
  if objective_goal == vz.ObjectiveMetricGoal.MAXIMIZE:
    alternative = 'less'
  else:
    alternative = 'greater'

  if len(candidate_mean_values) == 1:
    return stats.ttest_1samp(
        a=baseline_mean_values,
        popmean=candidate_mean_values[0],
        alternative=alternative).pvalue
  else:
    # use Welch’s t-test
    return stats.ttest_ind(
        baseline_mean_values,
        candidate_mean_values,
        equal_var=False,
        alternative=alternative).pvalue



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Stateful algorithm suggesters with datastores to be used in benchmarks.

Definitions:
------------
- A 'suggester' generates suggestions and update its state based on the provided
suggestion results.

- A 'state' uses a suggester and experimenter to simulate a study. It calls
the suggester multiple times and evaluate the suggestions with the experimenter.
Note that the experimenter may be non-static.
"""

import abc
from typing import Optional, Protocol, Sequence

import attr
from vizier import algorithms as vza
from vizier import pythia
from vizier import pyvizier as vz
from vizier._src.algorithms.policies.designer_policy import InRamDesignerPolicy
from vizier._src.benchmarks.experimenters.experimenter import Experimenter
from vizier._src.benchmarks.experimenters.experimenter_factory import ExperimenterFactory


@attr.define
class PolicySuggester:
  """Wraps a Policy and a datastore/supporter as an algorithm.

  NOTE: Updates to datastore should interface directly with supporter API. The
  policy will simply use the dataset that is in the supporter. This is the most
  faithful replication of the datastore vs suggest interaction in prod. Note
  that the algorithm/policy is updated with new Trials upon the suggest call.

  Typical usage:
    suggester = PolicySuggester(policy, supporter)
    # Trial ids are assigned by the supporter.
    suggestions = suggestion.suggest(...)
    trials = evaluate_and_complete_trials(suggestions)
    suggester.supporter.AddTrials(trials)
  """

  _policy: pythia.Policy = attr.field()
  _local_supporter: pythia.InRamPolicySupporter = attr.field()

  def suggest(self, batch_size: Optional[int]) -> Sequence[vz.Trial]:
    return self._local_supporter.SuggestTrials(
        self._policy, count=batch_size or 1
    )

  @property
  def supporter(self) -> pythia.InRamPolicySupporter:
    return self._local_supporter

  @classmethod
  def from_designer_factory(
      cls,
      problem: vz.ProblemStatement,
      designer_factory: vza.DesignerFactory[vza.Designer],
      supporter: Optional[pythia.InRamPolicySupporter] = None,
      seed: Optional[int] = None,
  ) -> 'PolicySuggester':
    """Initializes a PolicySuggester from Designer."""
    if supporter is None:
      supporter = pythia.InRamPolicySupporter(problem)
    # Wrap as policy and Policy
    policy = InRamDesignerPolicy(
        problem_statement=problem,
        supporter=supporter,
        designer_factory=designer_factory,
        seed=seed,
    )
    return PolicySuggester(policy=policy, local_supporter=supporter)


@attr.define
class BenchmarkState:
  """State of a benchmark run. It is altered via benchmark protocols."""

  experimenter: Experimenter
  algorithm: PolicySuggester


@attr.define(frozen=True)
class BenchmarkStateFactory(abc.ABC):
  """Factory class to generate new BenchmarkState."""

  @abc.abstractmethod
  def __call__(self, seed: Optional[int] = None) -> BenchmarkState:
    """Creates a new instance of BenchmarkFactory."""
    pass


@attr.define(frozen=True)
class ExperimenterDesignerBenchmarkStateFactory(BenchmarkStateFactory):
  """Generate new BenchmarkState from exptr/designer factory."""

  experimenter_factory: ExperimenterFactory
  designer_factory: vza.DesignerFactory[vza.Designer]

  def __call__(self, seed: Optional[int] = None) -> BenchmarkState:
    """Create a BenchmarkState from experimenter and designer factory."""
    experimenter = self.experimenter_factory()
    factory = DesignerBenchmarkStateFactory(
        experimenter=experimenter, designer_factory=self.designer_factory
    )
    return factory(seed=seed)


@attr.define(frozen=True)
class DesignerBenchmarkStateFactory(BenchmarkStateFactory):
  """Factory class to generate new BenchmarkState from designer factory."""

  experimenter: Experimenter
  designer_factory: vza.DesignerFactory[vza.Designer]

  def __call__(self, seed: Optional[int] = None) -> BenchmarkState:
    """Create a BenchmarkState from designer factory."""
    problem = self.experimenter.problem_statement()

    return BenchmarkState(
        experimenter=self.experimenter,
        algorithm=PolicySuggester.from_designer_factory(
            problem=problem,
            designer_factory=self.designer_factory,
            seed=seed,
        ),
    )


class SeededPolicyFactory(Protocol):
  """Factory to generate a policy with a seed."""

  def __call__(self, problem: vz.ProblemStatement, seed: int) -> pythia.Policy:
    ...


@attr.define(frozen=True)
class PolicyBenchmarkStateFactory(BenchmarkStateFactory):
  """Factory class to generate new BenchmarkState from policy factory."""

  experimenter: Experimenter
  policy_factory: SeededPolicyFactory

  def __call__(self, seed: Optional[int] = None) -> BenchmarkState:
    """Create a BenchmarkState from policy factory."""
    problem = self.experimenter.problem_statement()
    return BenchmarkState(
        experimenter=self.experimenter,
        algorithm=PolicySuggester(
            policy=self.policy_factory(problem, seed),
            local_supporter=pythia.InRamPolicySupporter(problem),
        ),
    )



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Essential classes for defining a blackbox optimization problem.

Contains `ProblemStatement` and its components that are cross platform
compatible.
"""
import collections
from typing import Collection
import enum
from typing import Callable, Iterable, Iterator, List, Optional, Type, TypeVar, Union
import attr
import numpy as np
from vizier._src.pyvizier.shared import common
from vizier._src.pyvizier.shared import parameter_config

################### PyTypes ###################
ScaleType = parameter_config.ScaleType
ExternalType = parameter_config.ExternalType
# A sequence of possible internal parameter values.
MonotypeParameterSequence = parameter_config.MonotypeParameterSequence
_T = TypeVar('_T')


################### Helper Classes ###################
def _min_leq_max(instance: 'MetricInformation', _, value: float):
  if value > instance.max_value:
    raise ValueError(
        f'min_value={value} cannot exceed max_value={instance.max_value}.'
    )


def _max_geq_min(instance: 'MetricInformation', _, value: float):
  if value < instance.min_value:
    raise ValueError(
        f'min_value={instance.min_value} cannot exceed max_value={value}.'
    )


# Values should NEVER be removed from ObjectiveMetricGoal, only added.
class ObjectiveMetricGoal(enum.IntEnum):
  """Valid Values for MetricInformation.Goal."""

  MAXIMIZE = 1
  MINIMIZE = 2

  # pylint: disable=comparison-with-callable
  @property
  def is_maximize(self) -> bool:
    return self == self.MAXIMIZE

  @property
  def is_minimize(self) -> bool:
    return self == self.MINIMIZE


class MetricType(enum.Enum):
  """Type of the metric.

  OBJECTIVE: Objective to be maximized / minimized.
  SAFETY: Objective to be kept above / below a certain threshold.
  """

  OBJECTIVE = 'OBJECTIVE'
  SAFETY = 'SAFETY'  # Soft constraint

  # pylint: disable=comparison-with-callable
  @property
  def is_safety(self) -> bool:
    return self == MetricType.SAFETY

  @property
  def is_objective(self) -> bool:
    return self == MetricType.OBJECTIVE


@attr.define(frozen=False, init=True, slots=True)
class MetricInformation:
  """MetricInformation provides optimization metrics configuration."""

  # The name of this metric. An empty string is allowed for single-metric
  # optimizations.
  name: str = attr.field(
      init=True, default='', validator=attr.validators.instance_of(str)
  )

  goal: ObjectiveMetricGoal = attr.field(
      init=True,
      # pylint: disable=g-long-lambda
      converter=ObjectiveMetricGoal,
      validator=attr.validators.instance_of(ObjectiveMetricGoal),
      on_setattr=[attr.setters.convert, attr.setters.validate],
      kw_only=True,
  )

  # The following should be used to configure this as a safety metric.
  # safety_threshold must always be set (to a float) for safety metrics.
  safety_threshold: Optional[float] = attr.field(
      init=True,
      default=None,
      validator=attr.validators.optional(attr.validators.instance_of(float)),
      kw_only=True,
  )
  # Safety_std_threshold is DEPRECATED and is here for backward compatibility.
  # To configure how cautious the optimization should be, please defer to
  # desired_min_safe_trials_fraction. This corresponds to the allowed
  # probability threshold (as a function of the z-score) of
  # violating the safety_threshold.
  safety_std_threshold: Optional[float] = attr.field(
      init=True,
      default=None,
      validator=attr.validators.optional(attr.validators.instance_of(float)),
      kw_only=True,
  )
  # Desired minimum fraction of safe trials (over total number of trials)
  # that should be targeted by the algorithm during the entire duration of the
  # study (best effort). A value of 0.0 is the same as None. If unset, or set
  # to 0.0, then Vizier has no constraint on the number/fraction of unsafe
  # trials it can suggest.
  desired_min_safe_trials_fraction: Optional[float] = attr.field(
      init=True,
      default=None,
      validator=[
          attr.validators.optional(attr.validators.instance_of(float)),
          attr.validators.optional(attr.validators.le(1.0)),
          attr.validators.optional(attr.validators.ge(0.0)),
      ],
      kw_only=True,
  )

  # Minimum value of this metric can be optionally specified.
  min_value: float = attr.field(
      init=True,
      default=None,
      # FYI: Converter is applied before validator.
      converter=lambda x: float(x) if x is not None else -np.inf,
      validator=[attr.validators.instance_of(float), _min_leq_max],
      kw_only=True,
  )

  # Maximum value of this metric can be optionally specified.
  max_value: float = attr.field(
      init=True,
      default=None,
      # FYI: Converter is applied before validator.
      converter=lambda x: float(x) if x is not None else np.inf,
      validator=[attr.validators.instance_of(float), _max_geq_min],
      on_setattr=attr.setters.validate,
      kw_only=True,
  )

  def min_value_or(self, default_value_fn: Callable[[], float]) -> float:
    """Returns the minimum value if finite, or default_value_fn().

    Avoids the common pitfalls of using
      `metric.min_value or default_value`
    which would incorrectly use the default_value when min_value == 0, and
    requires default_value to have been computed.

    Args:
      default_value_fn: Default value if min_value is not finite. This function
        does not run at all if min_value is finite.
    """
    if np.isfinite(self.min_value):
      return self.min_value
    else:
      return default_value_fn()

  def max_value_or(self, default_value_fn: Callable[[], float]) -> float:
    """Returns the minimum value if finite, or default_value_fn().

    Avoids the common pitfalls of using
      `metric.max_value or default_value`
    which would incorrectly use the default_value when max_value == 0, and
    requires default_value to have been computed.

    Args:
      default_value_fn: Default value if max_value is not finite. This function
        does not run at all if max_value is configured.
    """
    if np.isfinite(self.max_value):
      return self.max_value
    else:
      return default_value_fn()

  @property
  def range(self) -> float:
    """Range of the metric. Can be infinite."""
    return self.max_value - self.min_value

  @property
  def type(self) -> MetricType:
    if self.safety_threshold is not None:
      return MetricType.SAFETY
    else:
      return MetricType.OBJECTIVE

  def flip_goal(self) -> 'MetricInformation':
    """Flips the goal in-place and returns the reference to self."""
    if self.goal == ObjectiveMetricGoal.MAXIMIZE:
      self.goal = ObjectiveMetricGoal.MINIMIZE
    else:
      self.goal = ObjectiveMetricGoal.MAXIMIZE
    return self


@attr.define(frozen=False, init=True, slots=True)
class MetricsConfig(Collection[MetricInformation]):
  """Container for metrics.

  Metric names should be unique.
  """

  _metrics: List[MetricInformation] = attr.ib(
      init=True,
      factory=list,
      converter=list,
      validator=attr.validators.deep_iterable(
          member_validator=attr.validators.instance_of(MetricInformation),
          iterable_validator=attr.validators.instance_of(Iterable),
      ),
  )

  def item(self) -> MetricInformation:
    if len(self._metrics) != 1:
      raise ValueError(
          'item() may only be called when there is exactly one '
          'metric (there are %d).'
          % len(self._metrics)
      )
    return self._metrics[0]

  def _assert_names_are_unique(self) -> None:
    counts = collections.Counter(metric.name for metric in self._metrics)
    if len(counts) != len(self._metrics):
      for name, count in counts.items():
        if count > 1:
          raise ValueError(f'Duplicate metric name: {name} in {self._metrics}')

  def __attrs_post_init__(self):
    self._assert_names_are_unique()

  def __iter__(self) -> Iterator[MetricInformation]:
    return iter(self._metrics)

  def __contains__(self, x: object) -> bool:
    return x in self._metrics

  def __len__(self) -> int:
    return len(self._metrics)

  def __add__(self, metrics: Iterable[MetricInformation]) -> 'MetricsConfig':
    return MetricsConfig(self._metrics + list(metrics))

  def of_type(
      self, include: Union[MetricType, Iterable[MetricType]]
  ) -> 'MetricsConfig':
    """Filters the Metrics by type."""
    if isinstance(include, MetricType):
      include = (include,)
    return MetricsConfig(m for m in self._metrics if m.type in include)

  def exclude_type(
      self, exclude: Union[MetricType, Iterable[MetricType]]
  ) -> 'MetricsConfig':
    """Filters out the Metrics by type."""
    if isinstance(exclude, MetricType):
      exclude = (exclude,)
    return MetricsConfig(m for m in self._metrics if m.type not in exclude)

  def append(self, metric: MetricInformation):
    self._metrics.append(metric)
    self._assert_names_are_unique()

  def extend(self, metrics: Iterable[MetricInformation]):
    for metric in metrics:
      self.append(metric)

  @property
  def is_single_objective(self) -> bool:
    """Returns True if only one objective metric is configured."""
    return len(self.of_type(MetricType.OBJECTIVE)) == 1

  @property
  def is_safety_metric(self) -> bool:
    """Returns True if at least one safety metric is configured."""
    return True if self.of_type(MetricType.SAFETY) else False


################### Main Class ###################
@attr.define(frozen=False, init=True, slots=True)
class ProblemStatement:
  """Defines a blackbox optimization problem.

  `ProblemStatement` contains the minimal information that defines the search
  problem. It is inherited by platform-specific classes that carry additional
  platform-specific configurations including which algorithm to use.

  Each of OSS, Vertex, and Google Vizier has their own implementations of
  `StudyConfig` that inherit from `ProblemStatement`.

  Pythia `Policy` interface uses `ProblemStatement` as opposed to `StudyConfig`
  so that the same algorithm code can be used across platforms.
  """

  search_space: parameter_config.SearchSpace = attr.ib(
      init=True,
      factory=parameter_config.SearchSpace,
      validator=attr.validators.instance_of(parameter_config.SearchSpace),
      on_setattr=[attr.setters.convert, attr.setters.validate],
  )

  metric_information: MetricsConfig = attr.ib(
      init=True,
      factory=MetricsConfig,
      converter=MetricsConfig,
      validator=attr.validators.instance_of(MetricsConfig),
      on_setattr=[attr.setters.convert, attr.setters.validate],
      kw_only=True,
  )

  metadata: common.Metadata = attr.field(
      init=True,
      kw_only=True,
      factory=common.Metadata,
      validator=attr.validators.instance_of(common.Metadata),
      on_setattr=[attr.setters.convert, attr.setters.validate],
  )

  @property
  def debug_info(self) -> str:
    return ''

  @classmethod
  def from_problem(cls: Type[_T], problem: 'ProblemStatement') -> _T:
    """Converts a ProblemStatement to a subclass instance.

    Note that this method is useful in subclasses but not so much in
    `ProblemStatement` itself. `ProblemStatement.from_problem` simply generates
    a (shallow) copy of `problem`.

    Args:
      problem:

    Returns:
      A subclass instance filled with shallow copies of `ProblemStatement`
      fields.
    """
    return cls(
        search_space=problem.search_space,
        metric_information=problem.metric_information,
        metadata=problem.metadata,
    )

  def to_problem(self) -> 'ProblemStatement':
    """Converts to a ProblemStatement which is the parent class of `self`.

    Note that this method is useful in subclasses but not so much in
    `ProblemStatement` itself. `ProblemStatement.to_problem` simply generates
    a (shallow) copy of `problem`.

    Returns:
      `ProblemStatement` filled with shallow copies of `self.
    """
    return ProblemStatement(
        search_space=self.search_space,
        metric_information=self.metric_information,
        metadata=self.metadata,
    )

  @property
  def is_single_objective(self) -> bool:
    """Returns True if only one objective metric is configured."""
    return self.metric_information.is_single_objective

  @property
  def single_objective_metric_name(self) -> Optional[str]:
    """Returns the name of the single-objective metric, if set.

    Returns:
      String: name of the single-objective metric.
      None: if this is not a single-objective study.
    """
    if self.is_single_objective:
      return self.metric_information.of_type(MetricType.OBJECTIVE).item().name
    return None

  @property
  def is_safety_metric(self) -> bool:
    """Returns True if at least one safety metric is configured."""
    return self.metric_information.is_safety_metric



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for comparator_runner."""

from typing import Optional, Sequence

import jax
from jax import numpy as jnp
import numpy as np
from vizier import algorithms as vza
from vizier import benchmarks
from vizier import pyvizier as vz
from vizier._src.algorithms.optimizers import vectorized_base as vb
from vizier._src.algorithms.testing import comparator_runner
from vizier._src.jax import types
from vizier.benchmarks import experimenters
from vizier.pyvizier import converters

from absl.testing import absltest
from absl.testing import parameterized


class FakeVectorizedStrategy(vb.VectorizedStrategy):
  """Dummy vectorized strategy to control convergence."""

  def __init__(
      self,
      converter: converters.TrialToModelInputConverter,
      good_value: float = 1.0,
      bad_value: float = 0.0,
      num_trial_to_converge: int = 0,
  ):
    self.converter = converter
    self.good_value = good_value
    self.bad_value = bad_value
    self.num_trial_to_converge = num_trial_to_converge
    self.num_trials_so_far = 0

  def init_state(
      self,
      seed: jax.Array,
      n_parallel: int = 1,
      *,
      prior_features: Optional[vb.VectorizedOptimizerInput] = None,
      prior_rewards: Optional[types.Array] = None,
  ) -> None:
    return

  def suggest(
      self,
      seed: jax.Array,
      state: None,
      n_parallel: int = 1,
  ) -> vb.VectorizedOptimizerInput:
    continuous_output_len = sum(
        [spec.num_dimensions for spec in self.converter.output_specs.continuous]
    )
    categorical_output_len = len(self.converter.output_specs.categorical)
    shape = (1, n_parallel, continuous_output_len)
    if self.num_trials_so_far < self.num_trial_to_converge:
      continuous = jnp.ones(shape) * self.bad_value
    else:
      continuous = jnp.ones(shape) * self.good_value
    return vb.VectorizedOptimizerInput(
        continuous=continuous,
        categorical=jnp.zeros(
            (1, n_parallel, categorical_output_len), dtype=types.INT_DTYPE
        ),
    )

  @property
  def suggestion_batch_size(self) -> int:
    return 1

  def update(
      self,
      seed: jax.Array,
      state: None,
      batch_features: vb.VectorizedOptimizerInput,
      batch_rewards: types.Array,
  ) -> None:
    pass


class FakeDesigner(vza.Designer):
  """Dummy designer to control convergence."""

  def __init__(
      self,
      search_space: vz.SearchSpace,
      *,
      good_value: float = 1.0,
      bad_value: float = 0.0,
      noise: float = 0.1,
      num_trial_to_converge: int = 0,
      seed: Optional[int] = None,
  ):
    self.search_space = search_space
    self.good_value = good_value
    self.bad_value = bad_value
    self.noise = noise
    self.num_trial_to_converge = num_trial_to_converge
    self.num_trials_so_far = 0

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    del all_active
    self.num_trials_so_far += len(completed.trials)

  def suggest(
      self, count: Optional[int] = None
  ) -> Sequence[vz.TrialSuggestion]:
    if self.num_trials_so_far < self.num_trial_to_converge:
      parameters = {
          param.name: self.bad_value + self.noise * np.random.uniform()
          for param in self.search_space.parameters
      }
    else:
      parameters = {
          param.name: self.good_value + self.noise * np.random.uniform()
          for param in self.search_space.parameters
      }
    return [vz.TrialSuggestion(parameters)]


class EfficiencyConvergenceTest(absltest.TestCase):

  def test_comparison(self):
    experimenter = experimenters.BBOBExperimenterFactory('Sphere', 3)()

    num_trials = 20

    def _baseline_designer(
        problem: vz.ProblemStatement, seed: Optional[int] = None
    ) -> vza.Designer:
      return FakeDesigner(
          problem.search_space,
          num_trial_to_converge=num_trials,
          good_value=0.0,
          bad_value=1.0,
          seed=seed,
      )

    def _good_designer(
        problem: vz.ProblemStatement, seed: Optional[int] = None
    ) -> vza.Designer:
      return FakeDesigner(
          problem.search_space,
          good_value=0.0,
          bad_value=1.0,
          num_trial_to_converge=int(num_trials / 4),
          seed=seed,
      )

    comparator = comparator_runner.EfficiencyComparisonTester(
        num_trials=num_trials, num_repeats=5
    )
    comparator.assert_better_efficiency(
        benchmarks.DesignerBenchmarkStateFactory(
            experimenter=experimenter, designer_factory=_good_designer
        ),
        benchmarks.DesignerBenchmarkStateFactory(
            experimenter=experimenter, designer_factory=_baseline_designer
        ),
        score_threshold=0.3,
    )

    # Test that our baseline is worse.
    with self.assertRaises(comparator_runner.FailedComparisonTestError):  # pylint: disable=g-error-prone-assert-raises
      comparator.assert_better_efficiency(
          benchmarks.DesignerBenchmarkStateFactory(
              experimenter=experimenter, designer_factory=_baseline_designer
          ),
          benchmarks.DesignerBenchmarkStateFactory(
              experimenter=experimenter, designer_factory=_good_designer
          ),
          score_threshold=-0.1,
      )


class SimpleRegretConvergenceRunnerTest(parameterized.TestCase):
  """Test suite for convergence runner."""

  def setUp(self):
    super(SimpleRegretConvergenceRunnerTest, self).setUp()
    self.experimenter = experimenters.BBOBExperimenterFactory('Sphere', 3)()
    self.converter = converters.TrialToModelInputConverter.from_problem(
        self.experimenter.problem_statement()
    )

  @parameterized.parameters(
      {
          'candidate_num_trials': 1,
          'candidate_x_value': 5.0,
          'goal': vz.ObjectiveMetricGoal.MAXIMIZE,
          'should_pass': True,
      },
      {
          'candidate_num_trials': 100,
          'candidate_x_value': 5.0,
          'goal': vz.ObjectiveMetricGoal.MAXIMIZE,
          'should_pass': True,
      },
      {
          'candidate_num_trials': 1,
          'candidate_x_value': 0.0,
          'goal': vz.ObjectiveMetricGoal.MAXIMIZE,
          'should_pass': False,
      },
      {
          'candidate_num_trials': 100,
          'candidate_x_value': 0.0,
          'goal': vz.ObjectiveMetricGoal.MAXIMIZE,
          'should_pass': False,
      },
      {
          'candidate_num_trials': 1,
          'candidate_x_value': 5.0,
          'goal': vz.ObjectiveMetricGoal.MINIMIZE,
          'should_pass': False,
      },
      {
          'candidate_num_trials': 100,
          'candidate_x_value': 5.0,
          'goal': vz.ObjectiveMetricGoal.MINIMIZE,
          'should_pass': False,
      },
      {
          'candidate_num_trials': 1,
          'candidate_x_value': 0.0,
          'goal': vz.ObjectiveMetricGoal.MINIMIZE,
          'should_pass': True,
      },
      {
          'candidate_num_trials': 100,
          'candidate_x_value': 0.0,
          'goal': vz.ObjectiveMetricGoal.MINIMIZE,
          'should_pass': True,
      },
  )
  def test_designer_convergence(
      self, candidate_num_trials, candidate_x_value, goal, should_pass
  ):
    def _better_designer_factory(problem, seed):
      return FakeDesigner(
          search_space=problem.search_space,
          good_value=candidate_x_value,
          bad_value=0.0,
          noise=0.0,
          seed=seed,
      )

    def _baseline_designer_factory(problem, seed):
      return FakeDesigner(
          search_space=problem.search_space,
          good_value=1.0,
          bad_value=1.0,
          noise=0.0,
          seed=seed,
      )

    baseline_benchmark_state_factory = benchmarks.DesignerBenchmarkStateFactory(
        experimenter=self.experimenter,
        designer_factory=_baseline_designer_factory,
    )

    candidate_benchmark_state_factory = (
        benchmarks.DesignerBenchmarkStateFactory(
            experimenter=self.experimenter,
            designer_factory=_better_designer_factory,
        )
    )

    simple_regret_test = comparator_runner.SimpleRegretComparisonTester(
        baseline_num_trials=1000,
        candidate_num_trials=candidate_num_trials,
        baseline_suggestion_batch_size=1,
        candidate_suggestion_batch_size=1,
        baseline_num_repeats=5,
        candidate_num_repeats=5,
        alpha=0.05,
        goal=goal,
    )

    if should_pass:
      simple_regret_test.assert_benchmark_state_better_simple_regret(
          baseline_benchmark_state_factory,
          candidate_benchmark_state_factory,
      )
    else:
      with self.assertRaises(  # pylint: disable=g-error-prone-assert-raises
          comparator_runner.FailedSimpleRegretConvergenceTestError
      ):
        simple_regret_test.assert_benchmark_state_better_simple_regret(
            baseline_benchmark_state_factory,
            candidate_benchmark_state_factory,
        )

  @parameterized.parameters(
      {
          'candidate_x_value': 5.0,
          'goal': vz.ObjectiveMetricGoal.MAXIMIZE,
          'should_pass': True,
      },
      {
          'candidate_x_value': 0.0,
          'goal': vz.ObjectiveMetricGoal.MAXIMIZE,
          'should_pass': False,
      },
      {
          'candidate_x_value': 5.0,
          'goal': vz.ObjectiveMetricGoal.MINIMIZE,
          'should_pass': False,
      },
      {
          'candidate_x_value': 0.0,
          'goal': vz.ObjectiveMetricGoal.MINIMIZE,
          'should_pass': True,
      },
  )
  def test_optimizer_convergence(self, candidate_x_value, goal, should_pass):
    score_fn = lambda x, _: jnp.sum(x.continuous.padded_array, axis=-1)
    simple_regret_test = comparator_runner.SimpleRegretComparisonTester(
        baseline_num_trials=100,
        candidate_num_trials=100,
        baseline_suggestion_batch_size=1,
        candidate_suggestion_batch_size=1,
        baseline_num_repeats=5,
        candidate_num_repeats=5,
        alpha=0.05,
        goal=goal,
    )

    # pylint: disable=unused-argument
    def _baseline_strategy_factory(converter, suggestion_batch_size):
      return FakeVectorizedStrategy(
          converter=converter,
          good_value=1.0,
          bad_value=1.0,
          num_trial_to_converge=0,
      )

      # pylint: disable=unused-argument

    def _candidate_strategy_factory(converter, suggestion_batch_size):
      return FakeVectorizedStrategy(
          converter=converter,
          good_value=candidate_x_value,
          bad_value=0.0,
          num_trial_to_converge=0,
      )

    if should_pass:
      simple_regret_test.assert_optimizer_better_simple_regret(
          self.converter,
          score_fn,
          _baseline_strategy_factory,
          _candidate_strategy_factory,
      )
    else:
      with self.assertRaises(  # pylint: disable=g-error-prone-assert-raises
          comparator_runner.FailedSimpleRegretConvergenceTestError
      ):
        simple_regret_test.assert_optimizer_better_simple_regret(
            self.converter,
            score_fn,
            _baseline_strategy_factory,
            _candidate_strategy_factory,
        )


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for meta learning designer."""
from typing import Optional

import attrs
import numpy as np
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import quasi_random
from vizier._src.algorithms.designers.eagle_strategy import eagle_strategy
from vizier._src.algorithms.designers.meta_learning import meta_learning

from absl.testing import absltest
from absl.testing import parameterized


MetaLearningDesigner = meta_learning.MetaLearningDesigner
MetaLearningConfig = meta_learning.MetaLearningConfig
MetaLearningState = meta_learning.MetaLearningState


def _eagle_designer_factory(
    problem: vz.ProblemStatement, seed: Optional[int], **kwargs
):
  """Creates an EagleStrategyDesigner with hyper-parameters and seed."""
  config = eagle_strategy.FireflyAlgorithmConfig()
  # Unpack the hyperparameters into the Eagle config class.
  for param_name, param_value in kwargs.items():
    if param_name not in attrs.asdict(config):
      raise ValueError(f"'{param_name}' is not in FireflyAlgorithmConfig!")
    setattr(config, param_name, param_value)
  return eagle_strategy.EagleStrategyDesigner(
      problem_statement=problem,
      seed=seed,
      config=config,
  )


def _quasirandom_designer_factory(
    problem: vz.ProblemStatement, seed: Optional[int] = None
):
  """Creates a QuasiRandomDesigner with seed."""
  return quasi_random.QuasiRandomDesigner(problem.search_space, seed=seed)


def meta_learning_suggest_update_loop(
    meta_learning_designer: MetaLearningDesigner,
    num_suggestions: int,
    batch_size: int,
):
  trial_id = 0
  for _ in range(num_suggestions // batch_size):
    suggestions = meta_learning_designer.suggest(batch_size)
    trials = []
    for i in range(batch_size):
      trial = suggestions[i].to_trial(trial_id)
      trial_id += 1
      # Complete the trial with additional metric to test 'tuned_metric'.
      trial.complete(
          vz.Measurement(
              metrics={
                  'objective': np.random.random(),
                  'secondary_metric': np.random.random(),
              }
          )
      )
      trials.append(trial)
    meta_learning_designer.update(
        vza.CompletedTrials(trials), vza.ActiveTrials()
    )


class MetaLearningDesignerTest(parameterized.TestCase):
  """Tests for AutoTunerDesinger."""

  def setUp(self):
    super().setUp()
    self.problem = vz.ProblemStatement()
    self.problem.search_space.root.add_float_param('x', 0.0, 15.0)
    self.problem.search_space.root.add_float_param('y', -5.0, 10.0)
    self.problem.search_space.root.add_categorical_param('c', ['a', 'b', 'c'])
    self.problem.metric_information.append(
        vz.MetricInformation(
            name='objective',
            goal=vz.ObjectiveMetricGoal.MAXIMIZE,
        )
    )
    self.meta_config = MetaLearningConfig(
        num_trials_per_tuning=10,
        tuning_min_num_trials=100,
        tuning_max_num_trials=500,
    )
    self.tuning_params = vz.SearchSpace()
    self.tuning_params.root.add_float_param(
        'visibility', 0.0, 10.0, default_value=2.22
    )
    self.tuning_params.root.add_float_param(
        'gravity', 0.0, 10.0, default_value=3.33
    )

  def test_initialize_designer(self):
    meta_learning_designer = MetaLearningDesigner(
        problem=self.problem,
        tuning_hyperparams=self.tuning_params,
        tuned_designer_factory=_eagle_designer_factory,
        meta_designer_factory=_quasirandom_designer_factory,
        config=self.meta_config,
    )
    self.assertEqual(
        meta_learning_designer._state, MetaLearningState.INITIALIZE
    )
    # type: ignore[attribute-error]  # pylint: disable=protected-access
    self.assertEqual(
        meta_learning_designer._curr_tuned_designer._config.visibility, 2.22
    )
    self.assertEqual(
        meta_learning_designer._curr_tuned_designer._config.gravity, 3.33
    )

  @parameterized.parameters([1, 5])
  def test_state_transitions(self, batch_size):
    meta_learning_designer = MetaLearningDesigner(
        problem=self.problem,
        tuning_hyperparams=self.tuning_params,
        tuned_designer_factory=_eagle_designer_factory,
        meta_designer_factory=_quasirandom_designer_factory,
        config=self.meta_config,
    )
    self.assertEqual(
        meta_learning_designer._state,
        MetaLearningState.INITIALIZE,
    )
    meta_learning_suggest_update_loop(meta_learning_designer, 50, batch_size)
    self.assertEqual(
        meta_learning_designer._state,
        MetaLearningState.INITIALIZE,
    )
    meta_learning_suggest_update_loop(meta_learning_designer, 200, batch_size)
    self.assertEqual(
        meta_learning_designer._state, meta_learning.MetaLearningState.TUNE
    )
    meta_learning_suggest_update_loop(meta_learning_designer, 600, batch_size)
    self.assertEqual(
        meta_learning_designer._state,
        MetaLearningState.USE_BEST_PARAMS,
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import copy

from absl import logging
import attr
import numpy as np
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import random
from vizier._src.algorithms.evolution import numpy_populations
from vizier._src.algorithms.testing import test_runners
from vizier.testing import test_studies

from absl.testing import absltest


class NumpyPopulationsTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    problem = vz.ProblemStatement(
        test_studies.flat_space_with_all_types(),
        metric_information=test_studies.metrics_all_unconstrained())
    converter = numpy_populations.PopulationConverter(
        problem.search_space, problem.metric_information)

    trials = test_runners.run_with_random_metrics(
        random.RandomDesigner(problem.search_space, seed=0),
        problem,
        iters=1,
        batch_size=10,
        validate_parameters=False)
    population = converter.to_population(trials)

    self._problem = problem
    self._converter = converter
    self._trials = trials
    self._population = population

  def test_convert_to_population(self):
    logging.info('Population: %s', self._population)
    self.assertLen(self._population, 10)

  def test_population_serialization(self):
    dumped = self._population.dump()
    redumped = self._population.recover(dumped).dump()
    self.assertEqual(dumped, redumped)

  def test_linf_mutation_no_state_modification(self):
    population_copy = copy.deepcopy(self._population)

    mutation = numpy_populations.LinfMutation()
    mutation.mutate(self._population, 0)

    population_copy_dict = attr.asdict(population_copy)
    population_dict = attr.asdict(self._population)
    for k, v in population_copy_dict.items():
      np.testing.assert_array_equal(v, population_dict[k], err_msg=k)

  def test_convert_from_offspring(self):
    """TODO."""

  def test_offspring_serialization(self):
    """TODO."""

  def test_uniform_random_sampler(self):
    """TODO."""

  def test_linf_mutation(self):
    """TODO."""


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for vizier.pyvizier.shared.context."""

from vizier._src.pyvizier.shared import context
from absl.testing import absltest


class ContextTest(absltest.TestCase):

  def testDefaultsNotShared(self):
    """Make sure default parameters are not shared between instances."""
    context1 = context.Context()
    context2 = context.Context()
    context1.parameters['x1'] = context.ParameterValue(5)
    self.assertEmpty(context2.parameters)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tuner implementation based on Open Source Vizier."""

from concurrent import futures
import os
import threading
from typing import Optional, Union

from absl import logging
import attr
import grpc
import portpicker
import pyglove as pg
from vizier import pythia
from vizier import pyvizier as vz
from vizier._src.pyglove import algorithms
from vizier._src.pyglove import backend
from vizier._src.pyglove import client
from vizier._src.pyglove import converters
from vizier._src.service import clients as pyvizier_clients
from vizier._src.service import constants
from vizier._src.service import policy_factory as service_policy_factory_lib
from vizier._src.service import pythia_service
from vizier._src.service import pythia_service_pb2_grpc
from vizier._src.service import resources
from vizier._src.service import service_policy_supporter
from vizier._src.service import stubs_util
from vizier._src.service import types as vizier_types
from vizier._src.service import vizier_client
from vizier.client import client_abc
from vizier.service import pyvizier as svz

from google.protobuf import empty_pb2

BuiltinAlgorithm = algorithms.BuiltinAlgorithm
ExpandedStudyName = client.ExpandedStudyName
PolicyCache = client.PolicyCache
StudyKey = client.StudyKey


class PyGlovePolicyFactory(pythia.PolicyFactory):
  """PolicyFactory for OSSVizierTuner."""

  def __init__(self, policy_cache: PolicyCache):
    self._policy_cache = policy_cache

  def __call__(
      self, problem_statement, algorithm, policy_supporter, study_name
  ):
    study_resource = resources.StudyResource.from_name(study_name)
    study_key = StudyKey(
        study_resource.owner_id, ExpandedStudyName(study_resource.study_id)
    )
    if study_key in self._policy_cache:
      logging.info(
          'StudyKey %s was found in cache. Using it as the policy.', study_key
      )
      return self._policy_cache[study_key]

    # Use default Vizier algorithms if not using PyGlove poliices.
    logging.info(
        'StudyKey %s was not found in cache. Using default policy factory.'
    )

    default_policy_factory = service_policy_factory_lib.DefaultPolicyFactory()
    return default_policy_factory(
        problem_statement, algorithm, policy_supporter, study_name
    )


@attr.define
class _VizierServices:
  """Vizier services hub.

  This class is intended to be used as a singleton and not directly exposed to
  the user. Instead, users should always call `vizier.pyglove.init` to set up
  the vizier services stub, which calls `_VizierServices.use_vizier_service` and
  `_VizierServices.set_pythia_port` underlying.
  """

  _vizier_endpoint: Optional[str] = None
  _vizier_service: Optional[vizier_types.VizierService] = None
  _pythia_port: Optional[int] = None
  _pythia_server: Optional[grpc.Server] = None
  _pythia_servicer: Optional[pythia_service.PythiaServicer] = None

  def reset_for_testing(self) -> None:
    """Resets the services for testing purpose."""
    self._vizier_endpoint = None
    self._vizier_service = None
    self._pythia_port = None
    self._pythia_server = None
    self._pythia_servicer = None

  def use_vizier_service(self, endpoint: Optional[None]) -> None:
    """Uses vizier service specified by endpoint."""
    endpoint = endpoint or constants.NO_ENDPOINT
    if self._vizier_endpoint is not None and self._vizier_endpoint != endpoint:
      raise ValueError(
          'Cannot use different vizier endpoints in the same process. '
          f'Previous={self._vizier_endpoint}, New={endpoint}.'
      )
    self._vizier_endpoint = endpoint
    if self._vizier_service is not None:
      return
    if endpoint != constants.NO_ENDPOINT:
      pyvizier_clients.environment_variables.server_endpoint = (
          self._vizier_endpoint
      )
    self._vizier_service = vizier_client.create_vizier_servicer_or_stub()

  @property
  def vizier_service(self) -> vizier_types.VizierService:
    """Returns current vizier service."""
    assert self._vizier_service, 'call `use_vizier_service` first.'
    return self._vizier_service

  def set_pythia_port(self, pythia_port: Optional[int] = None) -> None:
    if (
        self._pythia_port is not None
        and pythia_port is not None
        and self._pythia_port != pythia_port
    ):
      raise ValueError(
          'Cannot use different pythia ports in the same process. '
          f'Previous={self._pythia_port}, New={pythia_port}'
      )
    self._pythia_port = pythia_port or portpicker.pick_unused_port()

  @property
  def pythia_endpoint(self) -> str:
    """Returns the endpoint for Pythia service."""
    assert self._pythia_port is not None, 'Please call `set_pythia_port` first.'
    return f'{os.uname()[1]}:{self._pythia_port}'

  def start_pythia_service(self, policy_cache: PolicyCache) -> None:
    """Start pythia service with a policy cache."""
    if self._pythia_server is not None:
      return

    policy_factory = PyGlovePolicyFactory(policy_cache)
    self._pythia_servicer = pythia_service.PythiaServicer(
        self._vizier_service, policy_factory
    )
    self._pythia_server = grpc.server(futures.ThreadPoolExecutor(max_workers=1))
    pythia_service_pb2_grpc.add_PythiaServiceServicer_to_server(
        self._pythia_servicer, self._pythia_server
    )
    self._pythia_server.add_insecure_port(self.pythia_endpoint)
    self._pythia_server.start()


# Global vizier services hub.
_services = _VizierServices()


class _OSSVizierTuner(client.VizierTuner):
  """OSS Vizier tuner for pyglove."""

  def get_tuner_id(self, algorithm: pg.DNAGenerator) -> str:
    """See parent class."""
    del algorithm
    # We use hostname plus thread ID as tuner ID, so we could test
    # distributed tuning scenarios using multi-threading.
    return f'{threading.get_ident()}@{_services.pythia_endpoint}'

  def _start_pythia_service(self, policy_cache: PolicyCache) -> None:
    """See parent class."""
    _services.start_pythia_service(policy_cache)

  def load_prior_study(self, resource_name: str) -> client_abc.StudyInterface:
    """See parent class."""
    return pyvizier_clients.Study.from_resource_name(resource_name)

  @classmethod
  def load_study(
      cls, owner: str, name: ExpandedStudyName
  ) -> client_abc.StudyInterface:
    """See parent class."""
    return pyvizier_clients.Study.from_owner_and_id(owner, name)

  def _configure_algorithm(
      self, study_config: svz.StudyConfig, algorithm: pg.DNAGenerator
  ) -> None:
    """Configure algorithm for a study."""
    if isinstance(algorithm, algorithms.BuiltinAlgorithm):
      study_config.algorithm = algorithm.name
    else:
      study_config.algorithm = 'EXTERNAL_PYTHIA_SERVICE'
    study_config.pythia_endpoint = _services.pythia_endpoint

  def create_study(
      self,
      problem: vz.ProblemStatement,
      converter: converters.VizierConverter,
      owner: str,
      name: str,
      algorithm: pg.DNAGenerator,
      stopping_policy: Optional[pg.tuning.EarlyStoppingPolicy] = None,
  ) -> client_abc.StudyInterface:
    """See parent class."""
    study_config = svz.StudyConfig.from_problem(problem)
    if converter.vizier_conversion_error:
      study_config.observation_noise = svz.ObservationNoise.HIGH
    self._configure_algorithm(study_config, algorithm)
    logging.info(
        'Created OSS Vizier study with owner: %s, name: %s', owner, name
    )
    return pyvizier_clients.Study.from_study_config(
        study_config, owner=owner, study_id=name
    )

  def get_group_id(self, group_id: Union[None, int, str] = None) -> str:
    """See parent class."""
    if group_id is None:
      hostname = os.uname()[1]
      thread_id = threading.get_ident()
      return f'{thread_id}@{hostname}'
    elif isinstance(group_id, int):
      return f'group:{group_id}'
    elif isinstance(group_id, str):
      return group_id  # pytype: disable=bad-return-type

  def ping_tuner(self, tuner_id: str) -> bool:
    # We treat `tuner_id` as the Pythia endpoint.
    assert '@' in tuner_id, tuner_id
    pythia_endpoint = tuner_id.split('@')[1]
    try:
      stubs_util.create_pythia_server_stub(pythia_endpoint, timeout=3).Ping(
          empty_pb2.Empty()
      )
      return True
    except (grpc.RpcError, grpc.FutureTimeoutError):
      return False

  def pythia_supporter(
      self, study: client_abc.StudyInterface
  ) -> pythia.PolicySupporter:
    return service_policy_supporter.ServicePolicySupporter(
        study.resource_name, _services.vizier_service
    )

  def use_pythia_for_study(self, study: client_abc.StudyInterface) -> None:
    pythia_endpoint = _services.pythia_endpoint
    metadata = svz.StudyConfig.pythia_endpoint_metadata(pythia_endpoint)
    study.update_metadata(metadata)


def init(
    study_prefix: Optional[str] = None,
    vizier_endpoint: Optional[str] = None,
    pythia_port: Optional[int] = None,
) -> None:
  """Init OSS Vizier backend.

  Args:
    study_prefix: An optional string that will be used as the prefix for the
      study names created by `pg.sample` throughout the application. This allows
      users to change the study names across multiple runs of the same binary
      through this single venue, instead of modifying the `name` argument of
      every `pg.sample` invocation.
    vizier_endpoint: An optional string in format of <hostname>:<port>, as the
      Vizier service address to connect to. If None, an in-process Vizier
      service will be created for local tuning scenarios.
    pythia_port: An optional port used for hosting the Pythia service. If None,
      the port will be automatically picked.
  """
  _services.use_vizier_service(vizier_endpoint)
  _services.set_pythia_port(pythia_port)
  backend.VizierBackend.use_study_prefix(study_prefix)
  pg.tuning.set_default_backend('oss_vizier')


@pg.tuning.add_backend('oss_vizier')
class OSSVizierBackend(backend.VizierBackend):
  """PyGlove backend that uses OSS Vizier."""

  tuner_cls = _OSSVizierTuner



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Designer that maps unsafe Trials as infeasible."""

import copy
from typing import Optional, Sequence
from absl import logging
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier.pyvizier import multimetric


class UnsafeAsInfeasibleDesigner(vza.Designer):
  """Designer that maps unsafe Trials as infeasible."""

  def __init__(
      self,
      problem_statement: vz.ProblemStatement,
      designer_factory: vza.DesignerFactory[vza.Designer],
      *,
      seed: Optional[int] = None,
  ):
    """Init.

    Args:
      problem_statement:
      designer_factory:
      seed:
    """
    self._safety_checker = multimetric.SafetyChecker(
        problem_statement.metric_information
    )
    problem_statement_safety_removed = copy.deepcopy(problem_statement)
    problem_statement_safety_removed.metric_information = (
        problem_statement.metric_information.exclude_type(vz.MetricType.SAFETY)
    )
    self._designer = designer_factory(
        problem_statement_safety_removed, seed=seed
    )

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    are_safe = self._safety_checker.are_trials_safe(completed.trials)
    mutated_trials = []
    mutated_trial_count = 0
    for is_safe, trial in zip(are_safe, completed.trials):
      if is_safe:
        mutated_trials.append(trial)
      else:
        measurement = trial.final_measurement or vz.Measurement()
        mutated_trial = trial.complete(
            measurement, infeasibility_reason='unsafe', inplace=False
        )
        mutated_trial_count += 1
        mutated_trials.append(mutated_trial)

    logging.info(
        'Update designer with %d completed trials (%d converted from safe to'
        ' infeasible)',
        len(mutated_trials),
        mutated_trial_count,
    )
    self._designer.update(vza.CompletedTrials(mutated_trials), all_active)

  def suggest(
      self, count: Optional[int] = None
  ) -> Sequence[vz.TrialSuggestion]:
    """Make new suggestions.

    Args:
      count: Makes best effort to generate this many suggestions. If None,
        suggests as many as the algorithm wants.

    Returns:
      New suggestions.
    """
    return self._designer.suggest(count)



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Randomization related utils."""

import logging
from typing import Dict, List, TypeVar

import numpy as np
from vizier import pyvizier as vz

_T = TypeVar('_T')


def sample_uniform(rng: np.random.Generator, min_value=0, max_value=1) -> float:
  """Samples unifrom value and udpate key."""
  return float(rng.uniform(low=min_value, high=max_value))


def sample_bernoulli(
    rng: np.random.Generator,
    prob1: float,
    value1: _T = 0,
    value2: _T = 1,
) -> _T:
  """Samples value1 with probability prob1."""
  return value1 if rng.binomial(1, p=prob1) else value2


def sample_integer(
    rng: np.random.Generator,
    min_value: float,
    max_value: float,
) -> int:
  """Samples a random integer."""
  val = sample_uniform(rng, min_value, max_value)
  return round(val)


def sample_categorical(rng: np.random.Generator, categories: List[str]) -> str:
  """Samples a random categorical value."""
  return rng.choice(categories)


def sample_discrete(rng: np.random.Generator,
                    feasible_points: List[float]) -> float:
  """Samples random discrete value.

  To sample a discrete value we sample uniformly a decimal value between the
  minimum and maximum feasible points and returns the closest feasible point.

  Args:
    rng:
    feasible_points:

  Returns:
    The sampled feasible point and a new key.
  """
  min_value = min(feasible_points)
  max_value = max(feasible_points)
  value = sample_uniform(rng, min_value, max_value)
  closest_element = get_closest_element(feasible_points, value)
  return closest_element


def get_closest_element(array: List[float], value: float) -> float:
  """Finds closest element in array to value."""
  gaps = [abs(x - value) for x in array]
  closest_idx = min(enumerate(gaps), key=lambda x: x[1])[0]
  return array[closest_idx]


def _sample_value(
    rng: np.random.Generator,
    param_config: vz.ParameterConfig,
) -> vz.ParameterValueTypes:
  """Samples random value based on the parameter type."""
  if param_config.type == vz.ParameterType.CATEGORICAL:
    return sample_categorical(rng, param_config.feasible_values)
  elif param_config.type == vz.ParameterType.DISCRETE:
    return sample_discrete(rng, param_config.feasible_values)
  else:
    min_value, max_value = param_config.bounds
    if param_config.type == vz.ParameterType.INTEGER:
      return sample_integer(rng, min_value, max_value)
    elif param_config.type == vz.ParameterType.DOUBLE:
      return sample_uniform(rng, min_value, max_value)
    else:
      logging.error('Invalid parameter config type: %s; deafults to DOUBLE.',
                    param_config.type)
      return sample_uniform(rng, min_value, max_value)


def sample_parameters(rng: np.random.Generator,
                      search_space: vz.SearchSpace) -> vz.ParameterDict:
  """Randomly samples parameter values from the search space."""
  sampled_parameters: Dict[str, vz.ParameterValue] = {}
  parameter_configs: List[vz.ParameterConfig] = search_space.parameters

  for param_config in parameter_configs:
    sample_param_value = _sample_value(rng, param_config)
    sampled_parameters[param_config.name] = vz.ParameterValue(
        sample_param_value)

  return vz.ParameterDict(sampled_parameters)


def shuffle_list(rng: np.random.Generator, items: List[_T]) -> List[_T]:
  """Create a new list of shuffled items. Original list remains the same."""
  shuffled_indices = np.array(range(len(items)))
  rng.shuffle(shuffled_indices)
  shuffled_items = [items[i] for i in shuffled_indices]
  return shuffled_items



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Serializable interface."""

import abc
from typing import Type, TypeVar

from vizier import pyvizier as vz

_S = TypeVar('_S', bound='Serializable')


class DecodeError(Exception):
  """Parent class for errors during decoding."""
  pass


class HarmlessDecodeError(DecodeError):
  """Failed to load state from metadata, and the object was untouched."""


class FatalDecodeError(DecodeError):
  """Failed to load state from metadata, and the object became invalid."""


class PartiallySerializable(abc.ABC):
  """Partially serializable objects except initialization.

  As long as the object is created in the same manner (via __init__ or a
  factory @classmethod), `load()` recovers the exact same `State`.

  NOTE: `State` here refers to the behavioral state that can be verified by
  calling "public" (not prefixed with an underline) methods. Subclasses do not
  guarantee that all private variables are recovered precisely.


  class Foo(PartiallySerializable):
   ...

  foo = Foo(*init_args)
  # do something with foo
  ...
  md = foo.dump()

  # `foo` and `foo2` are indistinguishable using "public" methods of Foo.
  foo2 = Foo(*init_args).load(md)

  # `foo` and `foo3` can be arbitrarily different.
  foo3 = Foo(*different_init_args).load(md)
  """

  @abc.abstractmethod
  def load(self, metadata: vz.Metadata) -> None:
    """Recovers the object's state stored in metadata.

    Args:
      metadata:

    Raises:
      HarmlessDecodeError: The object is still valid and can be used.
      FatalDecodeError: The object became invalid due to failure during
        load. It should be discarded. Raising a FatalDecodeError is NOT
        recommended, unless doing so avoids a signifcant overhead in performance
        or code complexity.
    """
    pass

  @abc.abstractmethod
  def dump(self) -> vz.Metadata:
    pass


class Serializable(abc.ABC):
  """Objects that can be fully serialized.

  Compared to PartiallySerializable, Serializable.recover() is a _classmethod_
  not a regular method. The object's `State` in its entirety can be serialized
  and then recovered.

  NOTE: `State` here refers to the behavioral state that can be verified by
  calling "public" (not prefixed with an underline) methods. Subclasses do not
  guarantee that all private variables are recovered precisely.


  class Foo(Serializable):
   ...

  foo = Foo()
  # do something with foo
  ...
  md = foo.dump()
  foo2 = Foo.recover(md)

  Then `foo` and `foo2` are indistinguishable using "public" methods of Foo.
  """

  # NOTE: cannot declare it as an abstract method due to Pytype limitations.
  @classmethod
  def recover(cls: Type['_S'], metadata: vz.Metadata) -> '_S':
    """Creates a new Serializable object from metadata.

    Args:
      metadata:

    Raises:
      DecodeError: Object could not be recovered from the metadata.
    """
    raise NotImplementedError('')

  @abc.abstractmethod
  def dump(self) -> vz.Metadata:
    pass



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Test runners for algorithms."""
from typing import Any, Callable, List, Optional, Sequence


from absl import logging
import attr
import numpy as np
from vizier import algorithms as vza
from vizier import benchmarks
from vizier import pythia
from vizier import pyvizier as vz
from vizier._src.algorithms.policies.designer_policy import InRamDesignerPolicy


@attr.define
class RandomMetricsRunner:
  """Generates completed `Trial`s with random metrics.

  EXAMPLE: This method can be used for smoke testing a `Designer`.
  ```
  from vizier._src.algorithms.testing import test_runners
  from vizier.testing import test_studies
  from vizier import pyvizier as vz

  problem = vz.ProblemStatement(
      test_studies.flat_space_with_all_types(),
      [vz.MetricInformation(
          'objective',
          goal=vz.ObjectiveMetricGoal.MAXIMIZE)])
  test_runners.RandomMetricsRunner(
      problem, validate_parameters=True).run_designer(my_designer)
  ```

  EXAMPLE: This method can be used for generating a large number of trials
  to be used as a test dataset.
  ```
  # (Continued from the above code block)
  from vizier._src.algorithms.designers import random
  trials = test_runners.RandomMetricsRunner(problem).run_designer(
      random.RandomDesigner(problem.search_space),
      validate_parameters=False)
  ```

  run_* methods run a suggest-update loop, completing suggestions with
  random metric values and then return all generated trials.
  The random values are sampled from uniform distribution over the metric's
  specified range. If the minimum or maximum is not specified, it defaults to
  -10 and +10 respectively.

  Attributes:
    problem: Problem definition.
    iters: Number of suggest-update iterations.
    batch_size: Number of suggestions to ask in each suggest() call.
    seed: Random seed for generating metrics.
    verbose: Increase the verbosity to see more logs.
    validate_parameters: If True, check if the suggested trials are valid in the
      search space.
  """

  problem: vz.ProblemStatement = attr.field()
  iters: int = attr.field(default=5)
  batch_size: Optional[int] = attr.field(default=1, kw_only=True)
  seed: Any = attr.field(default=None, kw_only=True)
  verbose: int = attr.field(default=0, kw_only=True)
  validate_parameters: bool = attr.field(default=False, kw_only=True)

  def _run(self, algorithm: benchmarks.PolicySuggester) -> List[vz.Trial]:
    """Implementation of run methods."""
    rng = np.random.RandomState(self.seed)
    for it in range(self.iters):
      suggestions = algorithm.suggest(self.batch_size)
      if not suggestions:
        logging.info(
            (
                'Preemptively finished at iteration %s'
                'because designer returned nothing.'
            ),
            it,
        )
        break

      for suggestion in suggestions:
        if self.validate_parameters:
          self.problem.search_space.assert_contains(suggestion.parameters)
        measurement = vz.Measurement()
        for mi in self.problem.metric_information:
          measurement.metrics[mi.name] = rng.uniform(
              mi.min_value_or(lambda: -10.0), mi.max_value_or(lambda: 10.0)
          )
        # PolicySupporter owns PolicySuggester, which owns reference to the
        # suggestions. Mutating these trials here will affect future
        # suggestions from the algorithm.
        suggestion.complete(measurement)
      if self.verbose:
        logging.info(
            'At iteration %s, trials suggested and evaluated:\n%s',
            it,
            suggestions,
        )
    return algorithm.supporter.GetTrials()

  def run_policy(
      self, policy_factory: Callable[[pythia.PolicySupporter], pythia.Policy]
  ) -> List[vz.Trial]:
    """Run the policy generated by policy_factory."""
    supporter = pythia.InRamPolicySupporter(self.problem)
    policy = policy_factory(supporter)
    return self._run(benchmarks.PolicySuggester(policy, supporter))

  def run_designer(self, designer: vza.Designer) -> List[vz.Trial]:
    """Run the specified $designer."""
    supporter = pythia.InRamPolicySupporter(self.problem)
    policy = InRamDesignerPolicy(
        self.problem,
        supporter=supporter,
        designer_factory=lambda _, **kwargs: designer,
    )
    return self._run(benchmarks.PolicySuggester(policy, supporter))


def run_with_random_metrics(
    designer: vza.Designer,
    problem: vz.ProblemStatement,
   
    iters: int = 5,
    *,
    batch_size: Optional[int] = 1,
    seed: Any = None,
    verbose: int = 0,
    validate_parameters: bool = False,
) -> Sequence[vz.Trial]:
  """DO NOT USE. DEPRECATED. Use RandomMetricsRunner.run_designer()."""
  return RandomMetricsRunner(
      problem,
      iters,
      batch_size=batch_size,
      seed=seed,
      verbose=verbose,
      validate_parameters=validate_parameters,
  ).run_designer(designer)



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for categorical_experimenter."""

from absl.testing import parameterized
from vizier import pyvizier as vz
from vizier._src.benchmarks.experimenters import l1_categorical_experimenter

from absl.testing import absltest


class L1CategoricalExperimenterTest(parameterized.TestCase):
  """Tests for CategoricalExperimenter."""

  def test_evaluate_optimum(self):
    optimum = [0, 1]
    exptr = l1_categorical_experimenter.L1CategorialExperimenter(
        num_categories=[2, 2], optimum=optimum)
    suggestion = vz.Trial(parameters={'c0': '0', 'c1': '1'})
    exptr.evaluate([suggestion])
    self.assertEqual(suggestion.final_measurement.metrics['objective'].value, 0)

  def test_evaluate_non_optimum(self):
    optimum = [0, 1]
    exptr = l1_categorical_experimenter.L1CategorialExperimenter(
        num_categories=[2, 2], optimum=optimum)
    suggestion = vz.Trial(parameters={'c0': '1', 'c1': '0'})
    exptr.evaluate([suggestion])
    self.assertEqual(suggestion.final_measurement.metrics['objective'].value, 2)

  @parameterized.parameters({'num_categories': [10, 3, 2]},
                            {'num_categories': [10, 2, 10]},
                            {'num_categories': [10, 10, 12, 15, 2]},
                            {'num_categories': [10]})
  def test_generate_random_optimum(self, num_categories):
    exptr = l1_categorical_experimenter.L1CategorialExperimenter(
        num_categories=num_categories)
    self.assertLen(exptr._optimum, len(num_categories))
    for i, value in enumerate(exptr._optimum.values()):
      self.assertIsInstance(value, str)
      self.assertTrue(0 <= int(value) < num_categories[i])

  def test_optimal_trial(self):
    optimum = [9, 2, 1]
    exptr = l1_categorical_experimenter.L1CategorialExperimenter(
        num_categories=[10, 3, 2], optimum=optimum)
    self.assertEqual(exptr.optimal_trial.parameters['c0'].value, '9')
    self.assertEqual(exptr.optimal_trial.parameters['c1'].value, '2')
    self.assertEqual(exptr.optimal_trial.parameters['c2'].value, '1')

  def test_optimal_trial_validation(self):
    optimum = [9, 3, 1]
    with self.assertRaises(ValueError):
      l1_categorical_experimenter.L1CategorialExperimenter(
          num_categories=[10, 3, 2], optimum=optimum)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""NAS-Bench-201 Experimenter.

Note that 'NATS' is an alias of 201.

Currently we only consider the topology search space (TSS) as it is much more
frequently used, as opposed to the size search space (SSS) which checks
performance by varying depths.
"""

from typing import Sequence
import nats_bench

from vizier import pyvizier
from vizier._src.benchmarks.experimenters import experimenter


def _model_tss_spc(ops: Sequence[str], num_nodes: int) -> str:
  """Converts ops and nodes to a string format recognized by NASBENCH-201."""
  nodes, k = [], 0
  for i in range(1, num_nodes):
    xstrs = []
    for j in range(i):
      xstrs.append('{:}~{:}'.format(ops[k], j))
      k += 1
    nodes.append('|' + '|'.join(xstrs) + '|')
  return '+'.join(nodes)


class NASBench201Experimenter(experimenter.Experimenter):
  """NASBENCH-201."""

  def __init__(self,
               nasbench: nats_bench.NATStopology,
               datastr_str: str = 'cifar10',
               validation_set_reporting_epoch: int = 12):
    self._nasbench = nasbench
    self._dataset_str = datastr_str
    tss_raw_config = nats_bench.search_space_info('nats-bench', 'tss')

    self._allowed_ops = tss_raw_config['op_names']
    self._num_nodes = tss_raw_config['num_nodes']
    self._op_spots = 6

  @property
  def nasbench(self) -> nats_bench.NATStopology:
    return self._nasbench

  def _trial_to_model_spec(self, trial: pyvizier.Trial):
    ops = [
        trial.parameters['op_{}'.format(i)].value for i in range(self._op_spots)
    ]
    return _model_tss_spc(ops, self._num_nodes)

  def evaluate(self, suggestions: Sequence[pyvizier.Trial]):
    for trial in suggestions:
      spec = self._trial_to_model_spec(trial)
      valid_acc, latency, time_cost, total_time = self._nasbench.simulate_train_eval(
          arch=spec, dataset=self._dataset_str)
      trial.complete(
          pyvizier.Measurement(
              metrics={
                  'valid_acc': valid_acc,
                  'latency': latency,
                  'time_cost': time_cost,
                  'total_time': total_time
              }))

  def problem_statement(self) -> pyvizier.ProblemStatement:
    problem_statement = pyvizier.ProblemStatement()
    root = problem_statement.search_space.root
    for i in range(self._op_spots):
      root.add_categorical_param(
          name='op_{}'.format(i), feasible_values=self._allowed_ops)
    problem_statement.metric_information.append(
        pyvizier.MetricInformation(
            name='valid_acc', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE))
    return problem_statement



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for hpob."""

from absl import app
from absl import logging

from vizier._src.benchmarks.experimenters import hpob_experimenter
from vizier._src.benchmarks.experimenters.hpob import handler as handler_lib

from absl.testing import absltest
from absl.testing import parameterized

# Global variable to be filled in with the class (not an instance) that shall
# be executed as the actual test. The class will be generated by
# generate_test_class().
class_to_test = None


def generate_test_class():
  """Generates all test cases. Must be called after InitGoogle()."""

  handler = handler_lib.HPOBHandler(
      root_dir=hpob_experimenter.ROOT_DIR,
      mode=hpob_experimenter.DEFAULT_TEST_MODE,
      surrogates_dir=hpob_experimenter.SURROGATES_DIR)

  class HpobTest(parameterized.TestCase):

    @absltest.skip("Files must be installed manually.")
    @parameterized.parameters(
        dict(normalize_y=True), dict(normalize_y=False),
        dict(normalize_y=True, use_surrogate_values=True),
        dict(normalize_y=False, use_surrogate_values=True),
        *tuple(dict(na_policy=x) for x in hpob_experimenter.NaPolicy))
    def test_all(self, **kwargs):
      container = hpob_experimenter.HPOBContainer(handler, **kwargs)
      for search_space_id, dataset_id in container.dataset_keys(
          hpob_experimenter.TEST):
        problem_and_trials = container.get_problem_and_trials(
            search_space_id, dataset_id)
        experimenter = container.get_experimenter(search_space_id, dataset_id)
        _ = experimenter.problem_statement()
        # TODO: Add more validations on the problem statement.
        objective = experimenter.EvaluateContinuous(
            problem_and_trials.trials[0])

        experimenter.evaluate([problem_and_trials.trials[0]])

        logging.info('First five trial metrics: %s', [
            t.final_measurement.metrics[hpob_experimenter.METRIC_NAME].value
            for t in problem_and_trials.trials[:5]
        ])
        logging.info('objective: %f', objective)
        logging.info('evaluated: %s', problem_and_trials.trials[0])

        self.assertAlmostEqual(
            problem_and_trials.trials[0].final_measurement.metrics[
                hpob_experimenter.METRIC_NAME].value,
            objective,
            places=1)

        break

  global class_to_test
  class_to_test = HpobTest


if __name__ == '__main__':
  app.call_after_init(generate_test_class)
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Applies the noise function to each metric in final measurement."""

import copy
from typing import Sequence
import attrs
from vizier import pyvizier as vz
from vizier._src.benchmarks.experimenters import experimenter


@attrs.define
class SwitchExperimenter(experimenter.Experimenter):
  """Creates conditional search space from multiple experimenters via a switch."""

  experimenters: Sequence[experimenter.Experimenter] = attrs.field()

  _switch_param_name: str = attrs.field(default='switch')
  _metric_name: str = attrs.field(default='switch_metric')

  # Created in __attrs_post_init__.
  _exptr_problems: Sequence[vz.ProblemStatement] = attrs.field(init=False)
  _exptr_objective_names: Sequence[str] = attrs.field(init=False)

  def __attrs_post_init__(self):
    self._exptr_problems = [
        exp.problem_statement() for exp in self.experimenters
    ]
    self._exptr_objective_names = [
        ps.metric_information.item().name for ps in self._exptr_problems
    ]

  def evaluate(self, suggestions: Sequence[vz.Trial]) -> None:
    for trial in suggestions:
      exptr_index = trial.parameters[self._switch_param_name].value

      trial_copy = copy.deepcopy(trial)
      self.experimenters[exptr_index].evaluate([trial_copy])

      if trial_copy.final_measurement is None:
        continue

      val = trial_copy.final_measurement.metrics[
          self._exptr_objective_names[exptr_index]
      ]
      trial.complete(vz.Measurement(metrics={self._metric_name: val}))

  def problem_statement(self) -> vz.ProblemStatement:
    problem_statement = vz.ProblemStatement()

    children = []
    for i, problem in enumerate(self._exptr_problems):
      for pc in problem.search_space.parameters:
        children.append(([i], pc))

    switching_param = vz.ParameterConfig.factory(
        self._switch_param_name,
        feasible_values=range(len(self.experimenters)),
        children=children,
    )
    problem_statement.search_space.add(switching_param)
    problem_statement.metric_information.append(
        vz.MetricInformation(
            name=self._metric_name, goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )

    return problem_statement



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Multiobjective Experimenters."""

import copy
from typing import Dict, Sequence


from vizier import pyvizier
from vizier._src.benchmarks.experimenters import experimenter


class MultiObjectiveExperimenter(experimenter.Experimenter):
  """MultiObjective Experimenter from a Dict of Experimenters."""

  def __init__(
      self,
      exptrs: Dict[str, experimenter.Experimenter],
  ):
    """Creates a MultiObjectiveExperimenter from a Dict of Experimenters.

    Specifically, the Experimenter takes a Dict of metric names and
    single-objective experimenters with the same search space. Each metric
    is evaluated individually by exptr and combined into a final measurement
    whose name is updated to metric name in exptrs. The problem statement
    of this experimenter is multi-objective with objective names given by keys
    in exptrs.

    Args:
      exptrs: Dict of metric name to Experimenters

    Raises:
      ValueError: Mismatching search space or exptr is not single-objective.
    """
    self._exptrs = exptrs

    # Copy and check that problem statements have same search space.
    self._problem_statement = list(exptrs.values())[0].problem_statement()
    first_search_space = self._problem_statement.search_space
    for exptr in exptrs.values():
      if exptr.problem_statement().search_space != first_search_space:
        raise ValueError(
            'Search space must match for all objectives: \n'
            f'{first_search_space} does not match '
            f'{exptr.problem_statement().search_space}'
        )

    metric_infos = []
    # Keeps track of the underlying metric information name of each extpr.
    self._previous_names = {}
    for name, exptr in exptrs.items():
      metric_info = exptr.problem_statement().metric_information.item()
      self._previous_names[name] = metric_info.name
      metric_info.name = name
      metric_infos.append(metric_info)

    self._problem_statement.metric_information = pyvizier.MetricsConfig(
        metric_infos
    )

  def problem_statement(self) -> pyvizier.ProblemStatement:
    return copy.deepcopy(self._problem_statement)

  def evaluate(self, suggestions: Sequence[pyvizier.Trial]):
    suggestions_copy = copy.deepcopy(suggestions)
    measurements = [pyvizier.Measurement() for _ in suggestions]
    for name, exptr in self._exptrs.items():
      exptr.evaluate(suggestions_copy)
      previous_name = self._previous_names[name]
      for idx, copied in enumerate(suggestions_copy):
        measurement = measurements[idx]
        assert copied.final_measurement is not None
        measurement.metrics[name] = copied.final_measurement.metrics[
            previous_name
        ]

    for suggestion, measurement in zip(suggestions, measurements):
      suggestion.complete(measurement)

    return suggestions

  def __repr__(self):
    return (
        f'MultiObjectiveExperimenter with {len(self._exptrs)} exptrs:'
        f' {self._exptrs}'
    )



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Base class for all PythiaPolicies."""

import abc
from typing import Any, FrozenSet, Optional

import attr
from vizier._src.pyvizier.pythia import study
from vizier._src.pyvizier.shared import base_study_config
from vizier._src.pyvizier.shared import trial


def _is_positive(instance: Any, attribute: Any, value: Any):
  del instance, attribute
  if value <= 0:
    raise ValueError(f'value must be positive! given: {value}')


def _not_empty(instance: Any, attribute, value: str):
  del instance, attribute
  if not value:
    raise ValueError(f'value must be not nullable! given: {value}')


@attr.define
class EarlyStopDecision:
  """Stopping decision on a single trial.

  Attributes:
    id: Trial's id.
    reason: Explanations for how `should_stop` value was determined.
    should_stop:
    metadata:
    predicted_final_measurement: This is added to the Trial so that other
      algorithms can treat it as if it is the observed final measurement.
  """
  id: int = attr.ib(
      validator=[attr.validators.instance_of(int), _is_positive],
      on_setattr=attr.setters.validate)
  # TODO: Record this in DB even when `should_stop` == False.
  reason: str = attr.ib(
      validator=[attr.validators.instance_of(str), _not_empty],
      on_setattr=attr.setters.validate,
      converter=str)
  should_stop: bool = attr.ib(
      default=True,
      validator=[attr.validators.instance_of(bool)],
      on_setattr=attr.setters.validate)

  # TODO: Add a proper support for this in the service side.
  # NOTE: As of 2022Q3, the standard deviation field of Metrics in this value
  #   are ignored (i.e. $predicted_final_measurement.metrics[].std).
  predicted_final_measurement: Optional[trial.Measurement] = attr.ib(
      default=None,
      validator=attr.validators.optional(
          attr.validators.instance_of(trial.Measurement)
      ),
      on_setattr=attr.setters.validate,
  )


@attr.define
class EarlyStopDecisions:
  """This is the output of the Policy.early_stop() method.

  Attributes:
    decisions: For some Trials, a decision as to whether it should be stopped.
    metadata: Metadata that's associated with the Study or with already existing
      Trials.
  """

  decisions: list[EarlyStopDecision] = attr.field(
      factory=list,
      validator=attr.validators.deep_iterable(
          attr.validators.instance_of(EarlyStopDecision)),
      converter=list)

  metadata: trial.MetadataDelta = attr.field(
      default=attr.Factory(trial.MetadataDelta),
      validator=attr.validators.instance_of(trial.MetadataDelta),
  )


@attr.define
class EarlyStopRequest:
  """Early stopping request.

  Attributes:
    study_guid:
    trial_ids: Trials to be considered for stopping, or None meaning "all
      Trials". This is a hint; it is allowable to consider stopping more or
      fewer trials.
    study_config:
    checkpoint_dir: If the policy wishes to use a checkpoint, then this is the
      path to find one.
    max_trial_id: max(trial.id for all existing Trials in the Study)
  """
  _study_descriptor: study.StudyDescriptor = attr.field(
      kw_only=True, validator=attr.validators.instance_of(study.StudyDescriptor)
  )

  trial_ids: Optional[FrozenSet[int]] = attr.field(
      default=None,
      validator=lambda x, c, v: x is None or isinstance(x, FrozenSet),
      converter=lambda x: None if x is None else frozenset(x))

  checkpoint_dir: Optional[str] = attr.field(
      default=None,
      validator=attr.validators.optional(attr.validators.instance_of(str)))

  @property
  def study_guid(self) -> str:
    return self._study_descriptor.guid

  @property
  def study_config(self) -> base_study_config.ProblemStatement:
    return self._study_descriptor.config

  @property
  def max_trial_id(self) -> int:
    return self._study_descriptor.max_trial_id


@attr.define(init=True)
class SuggestDecision:
  """This is the output of the Policy.suggestion() method.

  Attributes:
    suggestions: Trials to be suggested to the user.
    metadata: Metadata that's associated with the Study or with already existing
      Trials.
  """

  suggestions: list[trial.TrialSuggestion] = attr.field(
      init=True,
      validator=attr.validators.deep_iterable(
          attr.validators.instance_of(trial.TrialSuggestion)
      ),
      converter=list,
  )

  metadata: trial.MetadataDelta = attr.field(
      init=True,
      default=attr.Factory(trial.MetadataDelta),
      validator=attr.validators.instance_of(trial.MetadataDelta),
  )


@attr.define
class SuggestRequest:
  """Suggestion Request.

  Attributes:
    study_descriptor: information about the Study.
    study_guid: Study id
    count: A recommendation for how many suggestions should be generated.
    study_config:
    checkpoint_dir: (If set) A system-provided directory where the policy can
      store a checkpoint.
    max_trial_id: max(trial.id for all existing Trials in the Study)
  """
  _study_descriptor: study.StudyDescriptor = attr.field(
      validator=attr.validators.instance_of(study.StudyDescriptor),
      on_setattr=attr.setters.frozen,
      kw_only=True,
  )

  count: int = attr.field(
      validator=[attr.validators.instance_of(int), _is_positive],
      on_setattr=attr.setters.validate,
      kw_only=True)

  checkpoint_dir: Optional[str] = attr.field(
      default=None,
      validator=attr.validators.optional(attr.validators.instance_of(str)),
      on_setattr=attr.setters.validate,
      kw_only=True)

  @property
  def study_config(self) -> base_study_config.ProblemStatement:
    return self._study_descriptor.config

  @property
  def study_guid(self) -> str:
    return str(self._study_descriptor.guid)

  @property
  def max_trial_id(self) -> int:
    return self._study_descriptor.max_trial_id


class Policy(abc.ABC):
  """Interface for Pythia Policy subclasses.

  Most Policy subclasses would wish to take `PolicySupporter` object in the
  `__init__` function. `PolicySupporter` provides an abstraction for how
  `Policy` reads more information about the Study in question beyond the basic
  information available in `SuggestRequest` and `EarlyStopRequest`. It allows
  the `Policy` to be compatible in multiple environments.
  """

  @abc.abstractmethod
  def suggest(self, request: SuggestRequest) -> SuggestDecision:
    """Compute suggestions that Vizier will eventually hand to the user.

    Args:
      request:

    Returns:
      A bundle of TrialSuggestions and MetadataDelta that will be passed on to
        the user.  (See caveats in the SuggestionAnswer proto.)

    Raises:
      TemporaryPythiaError:  Generic retryable error.
      InactivateStudyError:  Raise this to inactivate the Study (non-retryable
        error).
        E.g. if this Policy cannot handle this StudyConfig.
        E.g. if this StudyConfig is somehow invalid.
        E.g. if this no more suggestions will ever be generated.
    """

  @abc.abstractmethod
  def early_stop(self, request: EarlyStopRequest) -> EarlyStopDecisions:
    """Decide which Trials Vizier should stop.

    This returns a list of decisions on on-going Trials.
    Args:
      request:

    Returns:
      Decisions about stopping some Trials and metadata changes.  If no decision
      is reported for a Trial, it's treated as "do not stop".

    Raises:
      TemporaryPythiaError:  Generic retryable error.
      InactivateStudyError: If this Pythia is inappropriate for the StudyConfig.
        (Non-retryable error.)  E.g. raise this if your Policy does not
        support MakeEarlyEarlyStopDecisions().
      CachedPolicyIsStaleError: Causes the computation to be restarted with a
        freshly constructed Policy instance.  It is incorrect to raise
        this on the first use of a Policy; the Study will be inactivated.
    """

  @property
  def name(self) -> str:
    """Returns the policy name. Intended for monitoring purposes only."""
    # Derived classes should override this implementation.
    return __class__.__name__

  @property
  def should_be_cached(self) -> bool:
    """Returns True if it's safe & worthwhile to cache this Policy in RAM.

    This is called after MakeEarlyEarlyStopDecisions() and/or MakeSuggestions().
    If True, the policy may be stored in RAM (at least for a while), and state
    may be preserved for the next time that Study makes it to that Pythia
    server.
    """
    return False



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Quasi-random designer."""

import collections
import math
import sys
import time
from typing import Optional, Sequence

import numpy as np
from scipy.stats import qmc
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier.pyvizier import converters


class QuasiRandomDesigner(vza.PartiallySerializableDesigner):
  """Sample points using quasi-random Halton algorithm."""

  def __init__(
      self,
      search_space: vz.SearchSpace,
      *,
      skip_points: int = 1000,
      seed: Optional[int] = None,
  ):
    """Init.

    Args:
      search_space: Must be a flat search space.
      skip_points: If positive, then these first points in the sequence are
        discarded in order to avoid unwanted correlations.
      seed: Random seed.
    """
    if search_space.is_conditional:
      raise ValueError(
          f'This designer {self} does not support conditional search.'
      )

    def _create_input_converter(pc):
      return converters.DefaultModelInputConverter(
          pc,
          scale=True,
          max_discrete_indices=sys.maxsize,
          float_dtype=np.float64,
      )

    self._converter = converters.DefaultTrialConverter(
        [_create_input_converter(pc) for pc in search_space.parameters]
    )

    for spec in self._converter.output_specs.values():
      if spec.type not in [
          converters.NumpyArraySpecType.CONTINUOUS,
          converters.NumpyArraySpecType.DISCRETE,
      ]:
        raise ValueError(f'Unsupported type: {spec.type} in {spec}')
      if spec.num_dimensions != 1:
        raise ValueError(
            'Multi-dimensional discrete types are unsuppored. Received spec: %s'
            % spec
        )
    self._seed = seed if seed is not None else np.int32(time.time())
    self._halton = qmc.Halton(
        d=len(self._converter.output_specs),
        seed=self._seed,
    )
    self._halton.fast_forward(skip_points)
    self._skip_points = skip_points
    self._output_specs = tuple(self._converter.output_specs.values())

  @classmethod
  def from_problem(
      cls, problem: vz.ProblemStatement, seed: Optional[int] = None
  ):
    """For wrapping via `PartiallySerializableDesignerPolicy`."""
    return QuasiRandomDesigner(problem.search_space, seed=seed)

  def load(self, metadata: vz.Metadata) -> None:
    """Loads designer's state from metadata."""
    self._seed = int(metadata.ns('quasi_random')['seed'])
    self._halton = qmc.Halton(
        d=len(self._converter.output_specs),
        seed=int(metadata.ns('quasi_random')['seed']),
    )
    self._skip_points = int(metadata.ns('quasi_random')['skip_points'])
    # Skip forward to where the previous Halton sequence stopped.
    self._halton.fast_forward(self._skip_points)

  def dump(self) -> vz.Metadata:
    """Dumps the designer's state."""
    metadata = vz.Metadata()
    metadata.ns('quasi_random')['skip_points'] = str(self._skip_points)
    metadata.ns('quasi_random')['seed'] = str(self._seed)
    return metadata

  def _generate_discrete_point(
      self, spec: converters.NumpyArraySpec, halton_value: float
  ) -> int:
    """Generate a discrete parameter value from a Halton value."""
    # +1 because the bounds are inclusive on both ends.
    num_discrete_options = spec.bounds[1] - spec.bounds[0] + 1 - spec.num_oovs
    # Get a number in [0,  num_discrete_options].
    halton_value *= num_discrete_options
    # Get an integer between 0 and num_discrete_options-1 (inclusive).
    halton_value = int(math.floor(halton_value))
    return halton_value + int(spec.bounds[0])

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    pass

  def suggest(
      self, count: Optional[int] = None
  ) -> Sequence[vz.TrialSuggestion]:
    """Suggest new suggestions, taking into account `count`."""
    count = count or 1
    sample = collections.defaultdict(list)
    for _ in range(count):
      # Generate random values in range of [0.0, 1.0].
      halton_values = self._halton.random(len(self._output_specs))[0]
      # Increment 'skip_points' to reflect designer's state.
      self._skip_points += len(self._output_specs)
      for dimension_index, spec in enumerate(self._output_specs):
        halton_value = halton_values[dimension_index]
        if spec.type == converters.NumpyArraySpecType.CONTINUOUS:
          # Samples are in [0,1]; Relies on the converter to scale back to
          # original scale once parameters are created.
          sample[spec.name].append(np.float64(halton_value))
        elif spec.type == converters.NumpyArraySpecType.DISCRETE:
          # Converter expects an integer for discrete/categorical parameters.
          sample[spec.name].append(
              np.int64(self._generate_discrete_point(spec, halton_value))
          )
        else:
          # Only CONTINUOUS and DISCRETE are supported (doesn't support ONEHOT).
          raise ValueError(
              f'Unsupported spec type: {spec.type}. {self._converter} should be'
              ' configured to return CONTINUOUS or DISCRETE specs only.'
          )
    sample = {
        name: np.expand_dims(np.asarray(elements), axis=-1)
        for (name, elements) in sample.items()
    }
    # Convert the samples back to parameters in the original search space.
    return [
        vz.TrialSuggestion(p) for p in self._converter.to_parameters(sample)
    ]



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Utility functions for performance tracking.

A typical use case is:

```
@jax.jit
@profiler.record_tracing
def _compute(self, x: jax.Array):
  return x + 1

class MyDesigner:

  @profiler.record_runtime
  def suggest(self) -> None:
    # An example of how to annotate an inline function call, so the function
    # name `_compute` is used. Note that the name must be specified.
    with profiler.timeit('compute'):
      _compute(jnp.array(1.))

with profiler.collect_events() as events:
  MyDesigner().suggest()
```

`events` is a list of `ProfileEvent`s. In this example, there are 3 events:
* Tracing of `_compute`, in scope `suggest::compute`
* Timing of `compute`, in scope `suggest`
* Timing of `suggest`, in empty scope.

`profiler.get_latencies_dict(events)` returns a dictionary whose values are
list of timedelta objects.

NOTE: @profiler.record_runtime and @profiler.record_tracing should appear
BEFORE the @jax.jit decorator, or only the first runtime will be recorded.
"""
import collections
import contextlib
import datetime
import enum
import functools
from typing import Any, Callable, Dict, Generator, List, Optional, Sequence

from absl import logging
import attrs
import jax


class EventType(enum.Enum):
  JIT_TRACING = 'tracing'
  TIMER = 'timer'


@attrs.define
class ProfileEvent:
  """An event logged by the profiler and stored in _Storage.

  Attributes:
    etype: type of the event.
    scope: context of the event.
    data: For JIT_TRACING event, it is either a singleton "function_name: str"
      OR a tuple of "function_name: str, args: tuple, kwargs: dict". For TIMER
      event, it's a tuple of "name: str" and "duration: timedelta".
  """

  etype: EventType = attrs.field()
  scope: str = attrs.field(converter='::'.join)
  data: Any = attrs.field()


@attrs.define
class _Storage:
  """Storage for function call runtimes.

  There is a single `_Storage` instance used globally. The global instance
  stays in "inactive" state until code enters a `collect_events()`
  contextmanager.

  Attributes:
    active: If not True, `add` method is a no-op.
    events:
    scope: Currently active "scope". Each time code enters
      `_enter_profile_scope` contextmanager, the scope name is appended to the
      list. Makes it easier to track where things happened.
  """

  active: bool = attrs.field(default=False)
  events: List[ProfileEvent] = attrs.field(factory=list)
  scope: list[str] = attrs.field(factory=list)

  def add(self, etype: EventType, data: Any) -> ProfileEvent:
    """Create and add a new event and return the added event.

    Args:
      etype: See ProfileEvent.
      data: See ProfileEvent.

    Returns:
      Newly added event.
    """
    event = ProfileEvent(etype, self.scope, data)
    if self.active:
      self.events.append(event)
    return event


_GLOBAL_SOTRAGE: _Storage = _Storage()


@contextlib.contextmanager
def _enter_profile_scope(scope: str):
  try:
    _GLOBAL_SOTRAGE.scope.append(scope)
    yield current_scope()
  finally:
    _GLOBAL_SOTRAGE.scope.pop()


def current_scope() -> str:
  return '::'.join(_GLOBAL_SOTRAGE.scope)


@contextlib.contextmanager
def collect_events() -> Generator[List[ProfileEvent], None, None]:
  """Context manager for turning on the profiler."""
  try:
    if _GLOBAL_SOTRAGE.active:
      raise RuntimeError(
          'There can be only one `collect_events()` context manager active at'
          ' the same time.'
      )
    _GLOBAL_SOTRAGE.active = True
    yield _GLOBAL_SOTRAGE.events
  finally:
    _GLOBAL_SOTRAGE.active = False
    # It's important to create a new instance here as opposed to resetting
    # the same object, so the yieled `events` persist outside the context.
    _GLOBAL_SOTRAGE.events = list()


@contextlib.contextmanager
def timeit(
    name: str, also_log: bool = False
) -> Generator[Callable[[], datetime.timedelta], None, None]:
  """Context manager for measuring the timing.

  Example:
  ```
  with timeit('scope_name') as duration:
    ...

  duration() # returns the duration.
  ```

  Also see: record_runtime, which is the decorator equivalent of this.

  Args:
    name:
    also_log: If True, also create a log.

  Yields:
    A callable with zero input arguments. Returns the elapsed time until call,
    or the end of the context, whichever comes earlier.
  """
  start = datetime.datetime.now()
  duration = None
  try:
    with _enter_profile_scope(name):

      def func() -> datetime.timedelta:
        if duration is None:
          return datetime.datetime.now() - start
        return duration

      yield func

  finally:
    duration = datetime.datetime.now() - start
    event = _GLOBAL_SOTRAGE.add(EventType.TIMER, (name, duration))
    if also_log:
      logging.info(
          'Timed %s: took %s seconds. Full event string: %s',
          name,
          duration.total_seconds(),
          event,
      )


def get_latencies_dict(
    events: Sequence[ProfileEvent],
) -> Dict[str, list[datetime.timedelta]]:
  d = collections.defaultdict(list)
  for e in events:
    if e.etype == EventType.TIMER:
      d[e.data[0]].append(e.data[1])
  return dict(d)


def record_runtime(
    func: Optional[Callable[..., Any]] = None,
    *,
    name_prefix: str = '',
    name: str = '',
    also_log: bool = False,
    block_until_ready: bool = False,
) -> Any:
  """Decorates the function to record the runtime.

  Also see: timeit(), which is the context manager equivalent of this.

  Args:
    func: Function being decorated.
    name_prefix: A prefix to add to the function name.
    name: The name to record. Defaults to func.__qualname__.
    also_log: Whether to also logging.info the runtime duration.
    block_until_ready: If True and running on an acceleartor, wait for the async
      dispatch results from this function, if any. See
      https://jax.readthedocs.io/en/latest/async_dispatch.html

  Returns:
    Decorated function, or decorator.
  """
  # This is required for the decorator to work both with and without
  # optional arguments.
  if func is None:
    return functools.partial(
        record_runtime,
        name_prefix=name_prefix,
        name=name,
        also_log=also_log,
        block_until_ready=block_until_ready,
    )
  name = name or func.__qualname__
  full_name = name
  if name_prefix:
    full_name = f'{name_prefix}.{name}'

  @functools.wraps(func)
  def wrapper(*args, **kwargs):
    """Calculates runtime of the given function."""
    with timeit(name=full_name, also_log=also_log):
      result = func(*args, **kwargs)
      if block_until_ready:
        result = jax.block_until_ready(result)
    return result
  return wrapper


@attrs.define
class _Config:
  """Configuration for this module."""

  _include_args_in_trace_records: bool = attrs.field(default=False)

  def include_args_in_trace_records(self, v: Optional[bool] = None) -> bool:
    """(Setter+getter): Whether to include input arguments in tracing events.

    NOTE: Use this feature for debugging only. If the function decorated with
    `@record_tracing` is not jitted, then arguments and kwargs will be actual
    Array(Tree)s instead of Tracers. The logs may include full values of the
    array which can greatly slow down your binary.

    Args:
      v: If None, leaves the behavior unchanged.

    Returns:
      New include_args_in_trace_records value.
    """
    if v is not None:
      self._include_args_in_trace_records = v
    return self._include_args_in_trace_records


config = _Config()


def record_tracing(
    func: Optional[Callable[..., Any]] = None,
    *,
    name: str = '',
    also_log: bool = True,
) -> Any:
  """Decorates the function to record the runtime of functions.

  Args:
    func: Function being decorated.
    name: The name to record. Defaults to func.__qualname__.
    also_log: Whether to also logging.info the runtime duration.

  Returns:
    Decorated function, or decorator.
  """
  # This is required for the decorator to work both with and without
  # optional arguments.
  if func is None:
    return functools.partial(
        record_tracing,
        name=name,
        also_log=also_log,
    )
  name = name or func.__qualname__

  @functools.wraps(func)
  def wrapper(*args, **kwargs):
    if config.include_args_in_trace_records():
      data = f'{name} with args={args} kwargs={kwargs}'
    else:
      data = name

    event = _GLOBAL_SOTRAGE.add(EventType.JIT_TRACING, data)
    if also_log:
      logging.info('Tracing %s. Full event string: %s', name, event)
    with _enter_profile_scope(name):
      result = func(*args, **kwargs)
    return result

  return wrapper



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Wrapper classes for Context protos and other messages in them."""
from typing import Dict, Optional

import attr
from vizier._src.pyvizier.shared import common
from vizier._src.pyvizier.shared import trial

Metadata = common.Metadata
ParameterValue = trial.ParameterValue


@attr.s(auto_attribs=True, frozen=False, init=True, slots=True)
class Context:
  """Wrapper for Context proto."""
  description: Optional[str] = attr.ib(
      init=True,
      kw_only=True,
      default=None,
      validator=attr.validators.optional(attr.validators.instance_of(str)),
      on_setattr=attr.setters.validate)

  parameters: Dict[str, ParameterValue] = attr.ib(
      init=True,
      kw_only=True,
      factory=dict,
      validator=attr.validators.deep_mapping(
          key_validator=attr.validators.instance_of(str),
          value_validator=attr.validators.instance_of(ParameterValue),
          mapping_validator=attr.validators.instance_of(dict)),
      on_setattr=attr.setters.validate)  # pytype: disable=wrong-arg-types

  metadata: Metadata = attr.ib(
      init=True,
      kw_only=True,
      default=Metadata(),
      validator=attr.validators.instance_of(Metadata),
      on_setattr=attr.setters.validate)

  related_links: Dict[str, str] = attr.ib(
      init=True,
      kw_only=True,
      factory=dict,
      validator=attr.validators.deep_mapping(
          key_validator=attr.validators.instance_of(str),
          value_validator=attr.validators.instance_of(str),
          mapping_validator=attr.validators.instance_of(dict)),
      on_setattr=attr.setters.validate)  # pytype: disable=wrong-arg-types



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Branin function."""
from typing import Sequence

import numpy as np
from vizier import pyvizier as vz
from vizier._src.benchmarks.experimenters import experimenter
from vizier._src.benchmarks.experimenters import numpy_experimenter


def _branin(x: np.ndarray) -> float:
  """Branin function.

  This function can accept batch shapes, although it is typed to return floats
  to conform to NumpyExperimenter API.

  Args:
    x: Shape (B*, 2) array.

  Returns:
    Shape (B*) array.
  """
  a = 1
  b = 5.1 / (4 * np.pi**2)
  c = 5 / np.pi
  r = 6
  s = 10
  t = 1 / (8 * np.pi)
  x1 = x[..., 0]
  x2 = x[..., 1]

  y = a * (x2 - b * x1**2 + c * x1 - r) ** 2 + s * (1 - t) * np.cos(x1) + s
  return y


class Branin2D(experimenter.Experimenter):
  """2D minimization function. See https://www.sfu.ca/~ssurjano/branin.html."""

  def __init__(self):
    self._impl = numpy_experimenter.NumpyExperimenter(
        _branin, self.problem_statement()
    )

  def evaluate(self, suggestions: Sequence[vz.Trial]) -> None:
    self._impl.evaluate(suggestions)

  def problem_statement(self) -> vz.ProblemStatement:
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param("x1", -5, 10)
    problem.search_space.root.add_float_param("x2", 0, 15)
    problem.metric_information.append(
        vz.MetricInformation(name="value", goal=vz.ObjectiveMetricGoal.MINIMIZE)
    )
    return problem



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from vizier import pyvizier
from vizier._src.benchmarks.experimenters import normalizing_experimenter
from vizier._src.benchmarks.experimenters import numpy_experimenter
from vizier._src.benchmarks.experimenters.synthetic import bbob

from absl.testing import absltest
from absl.testing import parameterized


class NormalizingExperimenterTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('Sphere', bbob.Sphere),
      ('Rastrigin', bbob.Rastrigin),
      ('BuecheRastrigin', bbob.BuecheRastrigin),
      ('LinearSlope', bbob.LinearSlope),
      ('AttractiveSector', bbob.AttractiveSector),
      ('StepEllipsoidal', bbob.StepEllipsoidal),
      ('RosenbrockRotated', bbob.RosenbrockRotated),
      ('Discus', bbob.Discus),
      ('BentCigar', bbob.BentCigar),
      ('SharpRidge', bbob.SharpRidge),
      ('DifferentPowers', bbob.DifferentPowers),
      ('Weierstrass', bbob.Weierstrass),
      ('SchaffersF7', bbob.SchaffersF7),
      ('SchaffersF7IllConditioned', bbob.SchaffersF7IllConditioned),
      ('GriewankRosenbrock', bbob.GriewankRosenbrock),
      ('Schwefel', bbob.Schwefel),
      ('Katsuura', bbob.Katsuura),
      ('Lunacek', bbob.Lunacek),
      ('Gallagher101Me', bbob.Gallagher101Me),
  )
  def testNormalizationApply(self, func):
    dim = 5
    exptr = numpy_experimenter.NumpyExperimenter(
        func, bbob.DefaultBBOBProblemStatement(dim)
    )
    normalizing_exptr = normalizing_experimenter.NormalizingExperimenter(
        exptr=exptr
    )

    parameters = exptr.problem_statement().search_space.parameters
    self.assertLen(parameters, dim)

    t = pyvizier.Trial(
        parameters={
            param.name: float(index) for index, param in enumerate(parameters)
        }
    )

    exptr.evaluate([t])
    metric_name = exptr.problem_statement().metric_information.item().name

    normalizing_exptr.evaluate([t])
    normalized_value = t.final_measurement.metrics[metric_name].value
    self.assertBetween(normalized_value, -10, 10)


if __name__ == '__main__':
  absltest.main()



# Copyright 2023 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for automated_stopping."""

from vizier._src.pyvizier.oss import automated_stopping
from vizier._src.service import study_pb2
from vizier._src.pyvizier.oss import compare
from absl.testing import absltest


class AutomatedStoppingTest(absltest.TestCase):

  def testDefaultStoppingConfig(self):
    config = automated_stopping.AutomatedStoppingConfig.default_stopping_spec()
    proto = study_pb2.StudySpec.DefaultEarlyStoppingSpec()
    compare.assertProto2Equal(self, proto, config.to_proto())


if __name__ == '__main__':
  absltest.main()
